# Copyright 2021 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime

from kubernetes_asyncio import client
from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.rest import ApiException
from kubernetes_asyncio.leaderelection.leaderelectionrecord import (
    LeaderElectionRecord,
)
from kubernetes_asyncio.leaderelection.resourcelock.baselock import BaseLock


class LeaseLock(BaseLock):
    def __init__(self, name: str, namespace: str, identity: str, api_client: ApiClient):
        """
        :param name: name of the lock
        :param namespace: namespace
        :param identity: A unique identifier that the candidate is using
        """
        super().__init__(name, namespace, identity)

        self.api_instance = client.CoordinationV1Api(api_client=api_client)

        # lease resource identity and reference
        self.lease_reference: client.V1Lease | None = None

    # get returns the election record from a Lease Annotation
    async def get(
        self, name: str, namespace: str
    ) -> tuple[bool, LeaderElectionRecord] | tuple[bool, Exception] | tuple[bool, None]:
        """
        :param name: Name of the lease object information to get
        :param namespace: Namespace in which the lease object is to be searched
        :return: 'True, election record' if object found else 'False, exception response'
        """
        try:
            lease = await self.api_instance.read_namespaced_lease(name, namespace)
        except ApiException as e:
            return False, e
        else:
            self.lease_reference = lease
            return True, self.election_record(lease)

    async def create(
        self, name: str, namespace: str, election_record: LeaderElectionRecord
    ) -> bool:
        """
        :param electionRecord: Annotation string
        :param name: Name of the lease object to be created
        :param namespace: Namespace in which the lease object is to be created
        :return: 'True' if object is created else 'False' if failed
        """
        body = client.V1Lease(
            metadata=client.V1ObjectMeta(name=name),
            spec=self.update_lease(election_record),
        )

        try:
            await self.api_instance.create_namespaced_lease(
                namespace, body, pretty=True
            )
            return True
        except ApiException:
            logging.exception("Failed to create lock")
            return False

    async def update(
        self, name: str, namespace: str, updated_record: LeaderElectionRecord
    ) -> bool:
        """
        :param name: name of the lock to be updated
        :param namespace: namespace the lock is in
        :param updated_record: the updated election record
        :return: True if update is successful False if it fails
        """
        try:
            # update the Lease from the updated record
            assert self.lease_reference is not None
            self.lease_reference.spec = self.update_lease(
                updated_record, self.lease_reference.spec
            )

            await self.api_instance.replace_namespaced_lease(
                name=name, namespace=namespace, body=self.lease_reference
            )
            return True
        except ApiException:
            logging.exception("Failed to update lock")
            return False

    def update_lease(
        self,
        leader_election_record: LeaderElectionRecord,
        current_spec: client.V1LeaseSpec | None = None,
    ):
        # existing or new lease?
        spec = current_spec if current_spec else client.V1LeaseSpec()

        # lease configuration
        assert leader_election_record.holder_identity
        spec.holder_identity = leader_election_record.holder_identity

        assert leader_election_record.lease_duration
        spec.lease_duration_seconds = int(leader_election_record.lease_duration)

        acquire_time = self.time_str_to_iso(leader_election_record.acquire_time)
        if acquire_time:
            spec.acquire_time = acquire_time

        renew_time = self.time_str_to_iso(leader_election_record.renew_time)
        if renew_time:
            spec.renew_time = renew_time

        return spec

    def election_record(self, lease: client.V1Lease):
        """
        Get leader election record from Lease spec.
        """
        leader_election_record = LeaderElectionRecord(None, None, None, None)

        if not lease.spec:
            return leader_election_record

        if lease.spec.holder_identity:
            leader_election_record.holder_identity = lease.spec.holder_identity
        if lease.spec.lease_duration_seconds:
            leader_election_record.lease_duration = str(
                lease.spec.lease_duration_seconds
            )
        if lease.spec.acquire_time:
            leader_election_record.acquire_time = str(
                datetime.replace(lease.spec.acquire_time, tzinfo=None)
            )
        if lease.spec.renew_time:
            leader_election_record.renew_time = str(
                datetime.replace(lease.spec.renew_time, tzinfo=None)
            )

        return leader_election_record

    # conversion between kubernetes ISO formatted time and elector record time
    def time_str_to_iso(self, str_time) -> datetime | None:
        formats = ["%Y-%m-%d %H:%M:%S.%f%z", "%Y-%m-%d %H:%M:%S.%f"]
        for fmt in formats:
            try:
                # return datetime.strptime(str_time, fmt).isoformat() + "Z"
                return datetime.strptime(str_time, fmt)
            except ValueError:
                pass
        logging.error("Failed to parse time string: %s", str_time)
        return None



from kubernetes_asyncio.client.models import V1PodFailurePolicyOnExitCodesRequirement
from kubernetes_asyncio.client.models import V1PodFailurePolicyOnPodConditionsPattern
from typing import Any

class V1PodFailurePolicyRule:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, action: str | None = None, on_exit_codes: V1PodFailurePolicyOnExitCodesRequirement | None = None, on_pod_conditions: list[V1PodFailurePolicyOnPodConditionsPattern] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def action(self) -> str: ...
    @action.setter
    def action(self, action: str) -> None: ...
    @property
    def on_exit_codes(self) -> V1PodFailurePolicyOnExitCodesRequirement: ...
    @on_exit_codes.setter
    def on_exit_codes(self, on_exit_codes: V1PodFailurePolicyOnExitCodesRequirement) -> None: ...
    @property
    def on_pod_conditions(self) -> list[V1PodFailurePolicyOnPodConditionsPattern]: ...
    @on_pod_conditions.setter
    def on_pod_conditions(self, on_pod_conditions: list[V1PodFailurePolicyOnPodConditionsPattern]) -> None: ...



from kubernetes_asyncio.client.models import V1ObjectMeta
from kubernetes_asyncio.client.models import V1PodSpec
from typing import Any

class V1PodTemplateSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, metadata: V1ObjectMeta | None = None, spec: V1PodSpec | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def metadata(self) -> V1ObjectMeta: ...
    @metadata.setter
    def metadata(self, metadata: V1ObjectMeta) -> None: ...
    @property
    def spec(self) -> V1PodSpec: ...
    @spec.setter
    def spec(self, spec: V1PodSpec) -> None: ...



from kubernetes_asyncio.client.models import V1IngressServiceBackend
from kubernetes_asyncio.client.models import V1TypedLocalObjectReference
from typing import Any

class V1IngressBackend:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, resource: V1TypedLocalObjectReference | None = None, service: V1IngressServiceBackend | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def resource(self) -> V1TypedLocalObjectReference: ...
    @resource.setter
    def resource(self, resource: V1TypedLocalObjectReference) -> None: ...
    @property
    def service(self) -> V1IngressServiceBackend: ...
    @service.setter
    def service(self, service: V1IngressServiceBackend) -> None: ...



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1.34.3
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes_asyncio.client.configuration import Configuration


class V1VolumeError(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'error_code': 'int',
        'message': 'str',
        'time': 'datetime'
    }

    attribute_map = {
        'error_code': 'errorCode',
        'message': 'message',
        'time': 'time'
    }

    def __init__(self, error_code=None, message=None, time=None, local_vars_configuration=None):  # noqa: E501
        """V1VolumeError - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default()
        self.local_vars_configuration = local_vars_configuration

        self._error_code = None
        self._message = None
        self._time = None
        self.discriminator = None

        if error_code is not None:
            self.error_code = error_code
        if message is not None:
            self.message = message
        if time is not None:
            self.time = time

    @property
    def error_code(self):
        """Gets the error_code of this V1VolumeError.  # noqa: E501

        errorCode is a numeric gRPC code representing the error encountered during Attach or Detach operations.  This is an optional, beta field that requires the MutableCSINodeAllocatableCount feature gate being enabled to be set.  # noqa: E501

        :return: The error_code of this V1VolumeError.  # noqa: E501
        :rtype: int
        """
        return self._error_code

    @error_code.setter
    def error_code(self, error_code):
        """Sets the error_code of this V1VolumeError.

        errorCode is a numeric gRPC code representing the error encountered during Attach or Detach operations.  This is an optional, beta field that requires the MutableCSINodeAllocatableCount feature gate being enabled to be set.  # noqa: E501

        :param error_code: The error_code of this V1VolumeError.  # noqa: E501
        :type error_code: int
        """

        self._error_code = error_code

    @property
    def message(self):
        """Gets the message of this V1VolumeError.  # noqa: E501

        message represents the error encountered during Attach or Detach operation. This string may be logged, so it should not contain sensitive information.  # noqa: E501

        :return: The message of this V1VolumeError.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this V1VolumeError.

        message represents the error encountered during Attach or Detach operation. This string may be logged, so it should not contain sensitive information.  # noqa: E501

        :param message: The message of this V1VolumeError.  # noqa: E501
        :type message: str
        """

        self._message = message

    @property
    def time(self):
        """Gets the time of this V1VolumeError.  # noqa: E501

        time represents the time the error was encountered.  # noqa: E501

        :return: The time of this V1VolumeError.  # noqa: E501
        :rtype: datetime
        """
        return self._time

    @time.setter
    def time(self, time):
        """Sets the time of this V1VolumeError.

        time represents the time the error was encountered.  # noqa: E501

        :param time: The time of this V1VolumeError.  # noqa: E501
        :type time: datetime
        """

        self._time = time

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1VolumeError):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1VolumeError):
            return True

        return self.to_dict() != other.to_dict()



from kubernetes_asyncio.client.models import V1VolumeProjection
from typing import Any

class V1ProjectedVolumeSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, default_mode: int | None = None, sources: list[V1VolumeProjection] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def default_mode(self) -> int: ...
    @default_mode.setter
    def default_mode(self, default_mode: int) -> None: ...
    @property
    def sources(self) -> list[V1VolumeProjection]: ...
    @sources.setter
    def sources(self, sources: list[V1VolumeProjection]) -> None: ...



from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.models import V1APIResourceList
from kubernetes_asyncio.client.models import V1CustomResourceDefinition
from kubernetes_asyncio.client.models import V1CustomResourceDefinitionList
from kubernetes_asyncio.client.models import V1Status
from multidict import CIMultiDictProxy
from typing import Any
from typing import Awaitable

class ApiextensionsV1Api:
    def __init__(self, api_client: ApiClient | None = None) -> None: ...
    def create_custom_resource_definition(self, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def create_custom_resource_definition_with_http_info(self, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def delete_collection_custom_resource_definition(self, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_collection_custom_resource_definition_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def delete_custom_resource_definition(self, name: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_custom_resource_definition_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def get_api_resources(self, **kwargs: Any) -> Awaitable[V1APIResourceList]: ...
    def get_api_resources_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1APIResourceList, int, CIMultiDictProxy]]: ...
    def list_custom_resource_definition(self, **kwargs: Any) -> Awaitable[V1CustomResourceDefinitionList]: ...
    def list_custom_resource_definition_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinitionList, int, CIMultiDictProxy]]: ...
    def patch_custom_resource_definition(self, name: str, body: Any, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def patch_custom_resource_definition_status(self, name: str, body: Any, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def patch_custom_resource_definition_status_with_http_info(self, name: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def patch_custom_resource_definition_with_http_info(self, name: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def read_custom_resource_definition(self, name: str, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def read_custom_resource_definition_status(self, name: str, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def read_custom_resource_definition_status_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def read_custom_resource_definition_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def replace_custom_resource_definition(self, name: str, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def replace_custom_resource_definition_status(self, name: str, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[V1CustomResourceDefinition]: ...
    def replace_custom_resource_definition_status_with_http_info(self, name: str, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...
    def replace_custom_resource_definition_with_http_info(self, name: str, body: V1CustomResourceDefinition, **kwargs: Any) -> Awaitable[tuple[V1CustomResourceDefinition, int, CIMultiDictProxy]]: ...



from kubernetes_asyncio.client.models import V1NodeSelectorRequirement
from typing import Any

class V1NodeSelectorTerm:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, match_expressions: list[V1NodeSelectorRequirement] | None = None, match_fields: list[V1NodeSelectorRequirement] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def match_expressions(self) -> list[V1NodeSelectorRequirement]: ...
    @match_expressions.setter
    def match_expressions(self, match_expressions: list[V1NodeSelectorRequirement]) -> None: ...
    @property
    def match_fields(self) -> list[V1NodeSelectorRequirement]: ...
    @match_fields.setter
    def match_fields(self, match_fields: list[V1NodeSelectorRequirement]) -> None: ...



from kubernetes_asyncio.client.models import V1AWSElasticBlockStoreVolumeSource
from kubernetes_asyncio.client.models import V1AzureDiskVolumeSource
from kubernetes_asyncio.client.models import V1AzureFilePersistentVolumeSource
from kubernetes_asyncio.client.models import V1CSIPersistentVolumeSource
from kubernetes_asyncio.client.models import V1CephFSPersistentVolumeSource
from kubernetes_asyncio.client.models import V1CinderPersistentVolumeSource
from kubernetes_asyncio.client.models import V1FCVolumeSource
from kubernetes_asyncio.client.models import V1FlexPersistentVolumeSource
from kubernetes_asyncio.client.models import V1FlockerVolumeSource
from kubernetes_asyncio.client.models import V1GCEPersistentDiskVolumeSource
from kubernetes_asyncio.client.models import V1GlusterfsPersistentVolumeSource
from kubernetes_asyncio.client.models import V1HostPathVolumeSource
from kubernetes_asyncio.client.models import V1ISCSIPersistentVolumeSource
from kubernetes_asyncio.client.models import V1LocalVolumeSource
from kubernetes_asyncio.client.models import V1NFSVolumeSource
from kubernetes_asyncio.client.models import V1ObjectReference
from kubernetes_asyncio.client.models import V1PhotonPersistentDiskVolumeSource
from kubernetes_asyncio.client.models import V1PortworxVolumeSource
from kubernetes_asyncio.client.models import V1QuobyteVolumeSource
from kubernetes_asyncio.client.models import V1RBDPersistentVolumeSource
from kubernetes_asyncio.client.models import V1ScaleIOPersistentVolumeSource
from kubernetes_asyncio.client.models import V1StorageOSPersistentVolumeSource
from kubernetes_asyncio.client.models import V1VolumeNodeAffinity
from kubernetes_asyncio.client.models import V1VsphereVirtualDiskVolumeSource
from typing import Any

class V1PersistentVolumeSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, access_modes: list[str] | None = None, aws_elastic_block_store: V1AWSElasticBlockStoreVolumeSource | None = None, azure_disk: V1AzureDiskVolumeSource | None = None, azure_file: V1AzureFilePersistentVolumeSource | None = None, capacity: dict[str, str] | None = None, cephfs: V1CephFSPersistentVolumeSource | None = None, cinder: V1CinderPersistentVolumeSource | None = None, claim_ref: V1ObjectReference | None = None, csi: V1CSIPersistentVolumeSource | None = None, fc: V1FCVolumeSource | None = None, flex_volume: V1FlexPersistentVolumeSource | None = None, flocker: V1FlockerVolumeSource | None = None, gce_persistent_disk: V1GCEPersistentDiskVolumeSource | None = None, glusterfs: V1GlusterfsPersistentVolumeSource | None = None, host_path: V1HostPathVolumeSource | None = None, iscsi: V1ISCSIPersistentVolumeSource | None = None, local: V1LocalVolumeSource | None = None, mount_options: list[str] | None = None, nfs: V1NFSVolumeSource | None = None, node_affinity: V1VolumeNodeAffinity | None = None, persistent_volume_reclaim_policy: str | None = None, photon_persistent_disk: V1PhotonPersistentDiskVolumeSource | None = None, portworx_volume: V1PortworxVolumeSource | None = None, quobyte: V1QuobyteVolumeSource | None = None, rbd: V1RBDPersistentVolumeSource | None = None, scale_io: V1ScaleIOPersistentVolumeSource | None = None, storage_class_name: str | None = None, storageos: V1StorageOSPersistentVolumeSource | None = None, volume_attributes_class_name: str | None = None, volume_mode: str | None = None, vsphere_volume: V1VsphereVirtualDiskVolumeSource | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def access_modes(self) -> list[str]: ...
    @access_modes.setter
    def access_modes(self, access_modes: list[str]) -> None: ...
    @property
    def aws_elastic_block_store(self) -> V1AWSElasticBlockStoreVolumeSource: ...
    @aws_elastic_block_store.setter
    def aws_elastic_block_store(self, aws_elastic_block_store: V1AWSElasticBlockStoreVolumeSource) -> None: ...
    @property
    def azure_disk(self) -> V1AzureDiskVolumeSource: ...
    @azure_disk.setter
    def azure_disk(self, azure_disk: V1AzureDiskVolumeSource) -> None: ...
    @property
    def azure_file(self) -> V1AzureFilePersistentVolumeSource: ...
    @azure_file.setter
    def azure_file(self, azure_file: V1AzureFilePersistentVolumeSource) -> None: ...
    @property
    def capacity(self) -> dict[str, str]: ...
    @capacity.setter
    def capacity(self, capacity: dict[str, str]) -> None: ...
    @property
    def cephfs(self) -> V1CephFSPersistentVolumeSource: ...
    @cephfs.setter
    def cephfs(self, cephfs: V1CephFSPersistentVolumeSource) -> None: ...
    @property
    def cinder(self) -> V1CinderPersistentVolumeSource: ...
    @cinder.setter
    def cinder(self, cinder: V1CinderPersistentVolumeSource) -> None: ...
    @property
    def claim_ref(self) -> V1ObjectReference: ...
    @claim_ref.setter
    def claim_ref(self, claim_ref: V1ObjectReference) -> None: ...
    @property
    def csi(self) -> V1CSIPersistentVolumeSource: ...
    @csi.setter
    def csi(self, csi: V1CSIPersistentVolumeSource) -> None: ...
    @property
    def fc(self) -> V1FCVolumeSource: ...
    @fc.setter
    def fc(self, fc: V1FCVolumeSource) -> None: ...
    @property
    def flex_volume(self) -> V1FlexPersistentVolumeSource: ...
    @flex_volume.setter
    def flex_volume(self, flex_volume: V1FlexPersistentVolumeSource) -> None: ...
    @property
    def flocker(self) -> V1FlockerVolumeSource: ...
    @flocker.setter
    def flocker(self, flocker: V1FlockerVolumeSource) -> None: ...
    @property
    def gce_persistent_disk(self) -> V1GCEPersistentDiskVolumeSource: ...
    @gce_persistent_disk.setter
    def gce_persistent_disk(self, gce_persistent_disk: V1GCEPersistentDiskVolumeSource) -> None: ...
    @property
    def glusterfs(self) -> V1GlusterfsPersistentVolumeSource: ...
    @glusterfs.setter
    def glusterfs(self, glusterfs: V1GlusterfsPersistentVolumeSource) -> None: ...
    @property
    def host_path(self) -> V1HostPathVolumeSource: ...
    @host_path.setter
    def host_path(self, host_path: V1HostPathVolumeSource) -> None: ...
    @property
    def iscsi(self) -> V1ISCSIPersistentVolumeSource: ...
    @iscsi.setter
    def iscsi(self, iscsi: V1ISCSIPersistentVolumeSource) -> None: ...
    @property
    def local(self) -> V1LocalVolumeSource: ...
    @local.setter
    def local(self, local: V1LocalVolumeSource) -> None: ...
    @property
    def mount_options(self) -> list[str]: ...
    @mount_options.setter
    def mount_options(self, mount_options: list[str]) -> None: ...
    @property
    def nfs(self) -> V1NFSVolumeSource: ...
    @nfs.setter
    def nfs(self, nfs: V1NFSVolumeSource) -> None: ...
    @property
    def node_affinity(self) -> V1VolumeNodeAffinity: ...
    @node_affinity.setter
    def node_affinity(self, node_affinity: V1VolumeNodeAffinity) -> None: ...
    @property
    def persistent_volume_reclaim_policy(self) -> str: ...
    @persistent_volume_reclaim_policy.setter
    def persistent_volume_reclaim_policy(self, persistent_volume_reclaim_policy: str) -> None: ...
    @property
    def photon_persistent_disk(self) -> V1PhotonPersistentDiskVolumeSource: ...
    @photon_persistent_disk.setter
    def photon_persistent_disk(self, photon_persistent_disk: V1PhotonPersistentDiskVolumeSource) -> None: ...
    @property
    def portworx_volume(self) -> V1PortworxVolumeSource: ...
    @portworx_volume.setter
    def portworx_volume(self, portworx_volume: V1PortworxVolumeSource) -> None: ...
    @property
    def quobyte(self) -> V1QuobyteVolumeSource: ...
    @quobyte.setter
    def quobyte(self, quobyte: V1QuobyteVolumeSource) -> None: ...
    @property
    def rbd(self) -> V1RBDPersistentVolumeSource: ...
    @rbd.setter
    def rbd(self, rbd: V1RBDPersistentVolumeSource) -> None: ...
    @property
    def scale_io(self) -> V1ScaleIOPersistentVolumeSource: ...
    @scale_io.setter
    def scale_io(self, scale_io: V1ScaleIOPersistentVolumeSource) -> None: ...
    @property
    def storage_class_name(self) -> str: ...
    @storage_class_name.setter
    def storage_class_name(self, storage_class_name: str) -> None: ...
    @property
    def storageos(self) -> V1StorageOSPersistentVolumeSource: ...
    @storageos.setter
    def storageos(self, storageos: V1StorageOSPersistentVolumeSource) -> None: ...
    @property
    def volume_attributes_class_name(self) -> str: ...
    @volume_attributes_class_name.setter
    def volume_attributes_class_name(self, volume_attributes_class_name: str) -> None: ...
    @property
    def volume_mode(self) -> str: ...
    @volume_mode.setter
    def volume_mode(self, volume_mode: str) -> None: ...
    @property
    def vsphere_volume(self) -> V1VsphereVirtualDiskVolumeSource: ...
    @vsphere_volume.setter
    def vsphere_volume(self, vsphere_volume: V1VsphereVirtualDiskVolumeSource) -> None: ...



from kubernetes_asyncio.client.models import V1PersistentVolumeSpec
from typing import Any

class V1VolumeAttachmentSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, inline_volume_spec: V1PersistentVolumeSpec | None = None, persistent_volume_name: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def inline_volume_spec(self) -> V1PersistentVolumeSpec: ...
    @inline_volume_spec.setter
    def inline_volume_spec(self, inline_volume_spec: V1PersistentVolumeSpec) -> None: ...
    @property
    def persistent_volume_name(self) -> str: ...
    @persistent_volume_name.setter
    def persistent_volume_name(self, persistent_volume_name: str) -> None: ...



from kubernetes_asyncio.client.models import V1IngressBackend
from kubernetes_asyncio.client.models import V1IngressRule
from kubernetes_asyncio.client.models import V1IngressTLS
from typing import Any

class V1IngressSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, default_backend: V1IngressBackend | None = None, ingress_class_name: str | None = None, rules: list[V1IngressRule] | None = None, tls: list[V1IngressTLS] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def default_backend(self) -> V1IngressBackend: ...
    @default_backend.setter
    def default_backend(self, default_backend: V1IngressBackend) -> None: ...
    @property
    def ingress_class_name(self) -> str: ...
    @ingress_class_name.setter
    def ingress_class_name(self, ingress_class_name: str) -> None: ...
    @property
    def rules(self) -> list[V1IngressRule]: ...
    @rules.setter
    def rules(self, rules: list[V1IngressRule]) -> None: ...
    @property
    def tls(self) -> list[V1IngressTLS]: ...
    @tls.setter
    def tls(self, tls: list[V1IngressTLS]) -> None: ...



# Copyright 2016 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
from typing import Any
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, Mock, call

import kubernetes_asyncio
from kubernetes_asyncio.client.exceptions import ApiException
from kubernetes_asyncio.watch import Watch


class WatchTest(IsolatedAsyncioTestCase):
    async def test_watch_with_decode(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.release = Mock()
        side_effects_events = [
            {
                "type": "ADDED",
                "object": {
                    "metadata": {
                        "name": "test{}".format(uid),
                        "resourceVersion": str(uid),
                    },
                    "spec": {},
                    "status": {},
                },
            }
            for uid in range(3)
        ]
        side_effects: list[Any] = [
            json.dumps(_).encode("utf8") for _ in side_effects_events
        ]
        side_effects.extend([AssertionError("Should not have been called")])
        fake_resp.content.readline.side_effect = side_effects

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        watch = kubernetes_asyncio.watch.Watch()
        count = 0
        async with watch:
            async for e in watch.stream(
                fake_api.get_namespaces, resource_version="123"
            ):
                self.assertEqual("ADDED", e["type"])
                # make sure decoder worked and we got a model with the right name
                self.assertEqual("test%d" % count, e["object"].metadata.name)
                # make sure decoder worked and updated Watch.resource_version
                self.assertEqual(e["object"].metadata.resource_version, str(count))
                self.assertEqual(watch.resource_version, str(count))

                # Stop the watch. This must not return the next event which would
                # be an AssertionError exception.
                count += 1
                if count == len(side_effects) - 1:
                    watch.stop()

        fake_api.get_namespaces.assert_called_once_with(
            _preload_content=False, watch=True, resource_version="123"
        )
        fake_resp.release.assert_called_once_with()

        # last resource_version has to be stored in the object
        self.assertEqual(watch.resource_version, "2")

    async def test_watch_for_follow(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.release = Mock()
        side_effects_logs = ["log_line_1", "log_line_2", ""]
        side_effects: list[Any] = [_.encode("utf8") for _ in side_effects_logs]
        side_effects.extend([AssertionError("Should not have been called")])
        fake_resp.content.readline.side_effect = side_effects

        fake_api = Mock()
        fake_api.read_namespaced_pod_log = AsyncMock(return_value=fake_resp)
        fake_api.read_namespaced_pod_log.__doc__ = (
            ":param follow:\n:type follow: bool\n:rtype: str"
        )

        watch = kubernetes_asyncio.watch.Watch()
        logs = []
        async with watch:
            async for e in watch.stream(fake_api.read_namespaced_pod_log):
                logs.append(e)

        self.assertListEqual(logs, ["log_line_1", "log_line_2"])
        fake_api.read_namespaced_pod_log.assert_called_once_with(
            _preload_content=False, follow=True
        )
        fake_resp.release.assert_called_once_with()

    async def test_watch_k8s_empty_response(self) -> None:
        """Stop the iterator when the response is empty.

        This typically happens when the user supplied timeout expires.

        """
        # Mock the readline return value to first return a valid response
        # followed by an empty response.
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        side_effects_events = [
            {
                "type": "ADDED",
                "object": {"metadata": {"name": "test0"}, "spec": {}, "status": {}},
            },
            {
                "type": "ADDED",
                "object": {"metadata": {"name": "test1"}, "spec": {}, "status": {}},
            },
        ]
        side_effects: list[Any] = [
            json.dumps(_).encode("utf8") for _ in side_effects_events
        ]
        fake_resp.content.readline.side_effect = side_effects + [b""]

        # Fake the K8s resource object to watch.
        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        # Iteration must cease after all valid responses were received.
        watch = kubernetes_asyncio.watch.Watch()
        cnt = 0
        async for _ in watch.stream(fake_api.get_namespaces):  # noqa
            cnt += 1
        self.assertEqual(cnt, len(side_effects))

    async def test_unmarshal_with_float_object(self) -> None:
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": 1}', "float")
        self.assertEqual("ADDED", event["type"])
        self.assertEqual(1.0, event["object"])
        self.assertTrue(isinstance(event["object"], float))
        self.assertEqual(1, event["raw_object"])

    async def test_unmarshal_without_return_type(self) -> None:
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": ["test1"]}', None)
        self.assertEqual("ADDED", event["type"])
        self.assertEqual(["test1"], event["object"])
        self.assertEqual(["test1"], event["raw_object"])

    async def test_unmarshal_with_empty_return_type(self) -> None:
        # empty string as a return_type is a default value
        # if watch can't detect object by function's name
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": ["test1"]}', "")
        self.assertEqual("ADDED", event["type"])
        self.assertEqual(["test1"], event["object"])
        self.assertEqual(["test1"], event["raw_object"])

    async def test_unmarshall_k8s_error_response(self) -> None:
        """Never parse messages of type ERROR.

        This test uses an actually recorded error, in this case for an outdated
        resource version.

        """
        # An actual error response sent by K8s during testing.
        k8s_err = {
            "type": "ERROR",
            "object": {
                "kind": "Status",
                "apiVersion": "v1",
                "metadata": {},
                "status": "Failure",
                "message": "too old resource version: 1 (8146471)",
                "reason": "Gone",
                "code": 410,
            },
        }

        with self.assertRaisesRegex(
            ApiException,
            r"\(410\)\nReason: Gone: too old resource version: 1 \(8146471\)",
        ):
            Watch().unmarshal_event(json.dumps(k8s_err), None)

    async def test_unmarshall_k8s_error_response_401_gke(self) -> None:
        """Never parse messages of type ERROR.

        This test uses an actually recorded error returned by GKE.

        """
        # An actual error response sent by K8s during testing.
        k8s_err = {
            "kind": "Status",
            "apiVersion": "v1",
            "metadata": {},
            "status": "Failure",
            "message": "Unauthorized",
            "reason": "Unauthorized",
            "code": 401,
        }

        with self.assertRaisesRegex(
            ApiException,
            r"\(401\)\nReason: Unauthorized: Unauthorized",
        ):
            Watch().unmarshal_event(json.dumps(k8s_err), None)

    async def test_unmarshal_with_custom_object(self) -> None:
        w = Watch()
        event = w.unmarshal_event(
            '{"type": "ADDED", "object": {"apiVersion":'
            '"test.com/v1beta1","kind":"foo","metadata":'
            '{"name": "bar", "resourceVersion": "1"}}}',
            "object",
        )
        self.assertEqual("ADDED", event["type"])
        # make sure decoder deserialized json into dictionary and updated
        # Watch.resource_version
        self.assertTrue(isinstance(event["object"], dict))
        self.assertEqual("1", event["object"]["metadata"]["resourceVersion"])
        self.assertEqual("1", w.resource_version)

    async def test_watch_with_exception(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.content.readline.side_effect = KeyError("expected")
        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        with self.assertRaises(KeyError):
            watch = kubernetes_asyncio.watch.Watch()
            async for e in watch.stream(fake_api.get_namespaces, timeout_seconds=10):  # noqa
                pass

    async def test_watch_retry_timeout(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.release = Mock()

        mock_event = {
            "type": "ADDED",
            "object": {
                "metadata": {"name": "test1555", "resourceVersion": "1555"},
                "spec": {},
                "status": {},
            },
        }

        fake_resp.content.readline.side_effect = [
            json.dumps(mock_event).encode("utf8"),
            asyncio.TimeoutError(),
            b"",
        ]

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        watch = kubernetes_asyncio.watch.Watch()
        async with watch.stream(fake_api.get_namespaces) as stream:
            async for e in stream:  # noqa
                pass

        fake_api.get_namespaces.assert_has_calls(
            [
                call(_preload_content=False, watch=True),
                call(_preload_content=False, watch=True, resource_version="1555"),
            ]
        )
        fake_resp.release.assert_called_once_with()

    async def test_watch_retry_410(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.release = Mock()

        mock_event1 = {
            "type": "ADDED",
            "object": {
                "metadata": {"name": "test1555", "resourceVersion": "1555"},
                "spec": {},
                "status": {},
            },
        }

        mock_event2 = {
            "type": "ADDED",
            "object": {
                "metadata": {"name": "test1555", "resourceVersion": "1555"},
                "spec": {},
                "status": {},
            },
        }

        mock_410 = {
            "type": "ERROR",
            "object": {
                "kind": "Status",
                "apiVersion": "v1",
                "metadata": {},
                "status": "Failure",
                "message": "too old resource version: 1 (8146471)",
                "reason": "Gone",
                "code": 410,
            },
        }

        # retry 410
        fake_resp.content.readline.side_effect = [
            json.dumps(mock_event1).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            json.dumps(mock_event2).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            b"",
        ]

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        watch = kubernetes_asyncio.watch.Watch()
        async with watch.stream(fake_api.get_namespaces) as stream:
            async for e in stream:  # noqa
                pass

        fake_api.get_namespaces.assert_has_calls(
            [
                call(_preload_content=False, watch=True),
                call(_preload_content=False, watch=True, resource_version="1555"),
            ]
        )
        fake_resp.release.assert_called_once_with()

        # retry 410 only once
        fake_resp.content.readline.side_effect = [
            json.dumps(mock_event1).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            json.dumps(mock_event2).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            b"",
        ]

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        with self.assertRaisesRegex(
            ApiException,
            r"\(410\)\nReason: Gone: too old resource version: 1 \(8146471\)",
        ):
            watch = kubernetes_asyncio.watch.Watch()
            async with watch.stream(fake_api.get_namespaces) as stream:
                async for e in stream:  # noqa
                    pass

        # no retry 410 if timeout is passed
        fake_resp.content.readline.side_effect = [
            json.dumps(mock_event1).encode("utf8"),
            json.dumps(mock_410).encode("utf8"),
            b"",
        ]

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        with self.assertRaisesRegex(
            ApiException,
            r"\(410\)\nReason: Gone: too old resource version: 1 \(8146471\)",
        ):
            watch = kubernetes_asyncio.watch.Watch()
            async with watch.stream(
                fake_api.get_namespaces, timeout_seconds=10
            ) as stream:
                async for e in stream:  # noqa
                    pass

    async def test_watch_timeout_with_resource_version(self) -> None:
        fake_resp = AsyncMock()
        fake_resp.content.readline = AsyncMock()
        fake_resp.release = Mock()

        fake_resp.content.readline.side_effect = [asyncio.TimeoutError(), b""]

        fake_api = Mock()
        fake_api.get_namespaces = AsyncMock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ":rtype: V1NamespaceList"

        watch = kubernetes_asyncio.watch.Watch()
        async with watch.stream(
            fake_api.get_namespaces, resource_version="10"
        ) as stream:
            async for e in stream:  # noqa
                pass

        # all calls use the passed resource version
        fake_api.get_namespaces.assert_has_calls(
            [
                call(_preload_content=False, watch=True, resource_version="10"),
                call(_preload_content=False, watch=True, resource_version="10"),
            ]
        )

        fake_resp.release.assert_called_once_with()
        self.assertEqual(watch.resource_version, "10")

    async def test_unmarshal_bookmark_succeeds_and_preserves_resource_version(
        self,
    ) -> None:
        w = Watch()
        event = w.unmarshal_event(
            '{"type": "BOOKMARK", "object": {"apiVersion":'
            '"test.com/v1beta1","kind":"foo","metadata":'
            '{"name": "bar", "resourceVersion": "1"}}}',
            "object",
        )
        self.assertEqual("BOOKMARK", event["type"])

        # make sure the resource version is preserved,
        # and the watcher's resource_version is updated
        self.assertTrue(isinstance(event["raw_object"], dict))
        self.assertEqual("1", event["raw_object"]["metadata"]["resourceVersion"])
        self.assertEqual("1", w.resource_version)

    async def test_unmarshal_job_bookmark_succeeds_and_preserves_resource_version(
        self,
    ) -> None:
        w = Watch()
        event = w.unmarshal_event(
            '{"type": "BOOKMARK", "object": {"apiVersion":'
            '"batch/v1","kind":"Job","metadata":'
            '{"name": "bar", "resourceVersion": "1"},'
            '"spec": {"template": {"metadata": '
            '{"creationTimestamp":null}, "spec": '
            '{"containers":null}}}}}',
            "object",
        )
        self.assertEqual("BOOKMARK", event["type"])

        # make sure the resource version is preserved,
        # and the watcher's resource_version is updated
        self.assertTrue(isinstance(event["raw_object"], dict))
        self.assertEqual("1", event["raw_object"]["metadata"]["resourceVersion"])
        self.assertEqual("1", w.resource_version)

    async def test_unmarshall_job_bookmark_malformed_object_fails(self) -> None:
        # An actual error response sent by K8s during testing.
        k8s_err = {
            "type": "BOOKMARK",
            "object": {
                "kind": "Job",
                "apiVersion": "batch/v1",
                "metadata": {},
            },
        }

        with self.assertRaises(Exception):
            Watch().unmarshal_event(json.dumps(k8s_err), None)



from kubernetes_asyncio.client.models import V1ScopeSelector
from typing import Any

class V1ResourceQuotaSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, hard: dict[str, str] | None = None, scope_selector: V1ScopeSelector | None = None, scopes: list[str] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def hard(self) -> dict[str, str]: ...
    @hard.setter
    def hard(self, hard: dict[str, str]) -> None: ...
    @property
    def scope_selector(self) -> V1ScopeSelector: ...
    @scope_selector.setter
    def scope_selector(self, scope_selector: V1ScopeSelector) -> None: ...
    @property
    def scopes(self) -> list[str]: ...
    @scopes.setter
    def scopes(self, scopes: list[str]) -> None: ...



from typing import Any

class V1QuobyteVolumeSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, group: str | None = None, read_only: bool | None = None, registry: str | None = None, tenant: str | None = None, user: str | None = None, volume: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def group(self) -> str: ...
    @group.setter
    def group(self, group: str) -> None: ...
    @property
    def read_only(self) -> bool: ...
    @read_only.setter
    def read_only(self, read_only: bool) -> None: ...
    @property
    def registry(self) -> str: ...
    @registry.setter
    def registry(self, registry: str) -> None: ...
    @property
    def tenant(self) -> str: ...
    @tenant.setter
    def tenant(self, tenant: str) -> None: ...
    @property
    def user(self) -> str: ...
    @user.setter
    def user(self, user: str) -> None: ...
    @property
    def volume(self) -> str: ...
    @volume.setter
    def volume(self, volume: str) -> None: ...



from datetime import datetime
from kubernetes_asyncio.client.models import V1ManagedFieldsEntry
from kubernetes_asyncio.client.models import V1OwnerReference
from typing import Any

class V1ObjectMeta:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, annotations: dict[str, str] | None = None, creation_timestamp: datetime | None = None, deletion_grace_period_seconds: int | None = None, deletion_timestamp: datetime | None = None, finalizers: list[str] | None = None, generate_name: str | None = None, generation: int | None = None, labels: dict[str, str] | None = None, managed_fields: list[V1ManagedFieldsEntry] | None = None, name: str | None = None, namespace: str | None = None, owner_references: list[V1OwnerReference] | None = None, resource_version: str | None = None, self_link: str | None = None, uid: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def annotations(self) -> dict[str, str]: ...
    @annotations.setter
    def annotations(self, annotations: dict[str, str]) -> None: ...
    @property
    def creation_timestamp(self) -> datetime: ...
    @creation_timestamp.setter
    def creation_timestamp(self, creation_timestamp: datetime) -> None: ...
    @property
    def deletion_grace_period_seconds(self) -> int: ...
    @deletion_grace_period_seconds.setter
    def deletion_grace_period_seconds(self, deletion_grace_period_seconds: int) -> None: ...
    @property
    def deletion_timestamp(self) -> datetime: ...
    @deletion_timestamp.setter
    def deletion_timestamp(self, deletion_timestamp: datetime) -> None: ...
    @property
    def finalizers(self) -> list[str]: ...
    @finalizers.setter
    def finalizers(self, finalizers: list[str]) -> None: ...
    @property
    def generate_name(self) -> str: ...
    @generate_name.setter
    def generate_name(self, generate_name: str) -> None: ...
    @property
    def generation(self) -> int: ...
    @generation.setter
    def generation(self, generation: int) -> None: ...
    @property
    def labels(self) -> dict[str, str]: ...
    @labels.setter
    def labels(self, labels: dict[str, str]) -> None: ...
    @property
    def managed_fields(self) -> list[V1ManagedFieldsEntry]: ...
    @managed_fields.setter
    def managed_fields(self, managed_fields: list[V1ManagedFieldsEntry]) -> None: ...
    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, name: str) -> None: ...
    @property
    def namespace(self) -> str: ...
    @namespace.setter
    def namespace(self, namespace: str) -> None: ...
    @property
    def owner_references(self) -> list[V1OwnerReference]: ...
    @owner_references.setter
    def owner_references(self, owner_references: list[V1OwnerReference]) -> None: ...
    @property
    def resource_version(self) -> str: ...
    @resource_version.setter
    def resource_version(self, resource_version: str) -> None: ...
    @property
    def self_link(self) -> str: ...
    @self_link.setter
    def self_link(self, self_link: str) -> None: ...
    @property
    def uid(self) -> str: ...
    @uid.setter
    def uid(self, uid: str) -> None: ...



from kubernetes_asyncio.client.models import V1HTTPHeader
from typing import Any

class V1HTTPGetAction:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, host: str | None = None, http_headers: list[V1HTTPHeader] | None = None, path: str | None = None, port: Any | None = None, scheme: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def host(self) -> str: ...
    @host.setter
    def host(self, host: str) -> None: ...
    @property
    def http_headers(self) -> list[V1HTTPHeader]: ...
    @http_headers.setter
    def http_headers(self, http_headers: list[V1HTTPHeader]) -> None: ...
    @property
    def path(self) -> str: ...
    @path.setter
    def path(self, path: str) -> None: ...
    @property
    def port(self) -> Any: ...
    @port.setter
    def port(self, port: Any) -> None: ...
    @property
    def scheme(self) -> str: ...
    @scheme.setter
    def scheme(self, scheme: str) -> None: ...



from typing import Any

class V1FileKeySelector:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, key: str | None = None, optional: bool | None = None, path: str | None = None, volume_name: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def key(self) -> str: ...
    @key.setter
    def key(self, key: str) -> None: ...
    @property
    def optional(self) -> bool: ...
    @optional.setter
    def optional(self, optional: bool) -> None: ...
    @property
    def path(self) -> str: ...
    @path.setter
    def path(self, path: str) -> None: ...
    @property
    def volume_name(self) -> str: ...
    @volume_name.setter
    def volume_name(self, volume_name: str) -> None: ...



from datetime import datetime
from typing import Any

class V1LeaseSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, acquire_time: datetime | None = None, holder_identity: str | None = None, lease_duration_seconds: int | None = None, lease_transitions: int | None = None, preferred_holder: str | None = None, renew_time: datetime | None = None, strategy: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def acquire_time(self) -> datetime: ...
    @acquire_time.setter
    def acquire_time(self, acquire_time: datetime) -> None: ...
    @property
    def holder_identity(self) -> str: ...
    @holder_identity.setter
    def holder_identity(self, holder_identity: str) -> None: ...
    @property
    def lease_duration_seconds(self) -> int: ...
    @lease_duration_seconds.setter
    def lease_duration_seconds(self, lease_duration_seconds: int) -> None: ...
    @property
    def lease_transitions(self) -> int: ...
    @lease_transitions.setter
    def lease_transitions(self, lease_transitions: int) -> None: ...
    @property
    def preferred_holder(self) -> str: ...
    @preferred_holder.setter
    def preferred_holder(self, preferred_holder: str) -> None: ...
    @property
    def renew_time(self) -> datetime: ...
    @renew_time.setter
    def renew_time(self, renew_time: datetime) -> None: ...
    @property
    def strategy(self) -> str: ...
    @strategy.setter
    def strategy(self, strategy: str) -> None: ...



from kubernetes_asyncio.client.models import V1ListMeta
from kubernetes_asyncio.client.models import V1alpha1PodCertificateRequest
from typing import Any

class V1alpha1PodCertificateRequestList:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, api_version: str | None = None, items: list[V1alpha1PodCertificateRequest] | None = None, kind: str | None = None, metadata: V1ListMeta | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def api_version(self) -> str: ...
    @api_version.setter
    def api_version(self, api_version: str) -> None: ...
    @property
    def items(self) -> list[V1alpha1PodCertificateRequest]: ...
    @items.setter
    def items(self, items: list[V1alpha1PodCertificateRequest]) -> None: ...
    @property
    def kind(self) -> str: ...
    @kind.setter
    def kind(self, kind: str) -> None: ...
    @property
    def metadata(self) -> V1ListMeta: ...
    @metadata.setter
    def metadata(self, metadata: V1ListMeta) -> None: ...



from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.models import VersionInfo
from multidict import CIMultiDictProxy
from typing import Any
from typing import Awaitable

class VersionApi:
    def __init__(self, api_client: ApiClient | None = None) -> None: ...
    def get_code(self, **kwargs: Any) -> Awaitable[VersionInfo]: ...
    def get_code_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[VersionInfo, int, CIMultiDictProxy]]: ...



from kubernetes_asyncio.client.models import V1ParentReference
from typing import Any

class V1IPAddressSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, parent_ref: V1ParentReference | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def parent_ref(self) -> V1ParentReference: ...
    @parent_ref.setter
    def parent_ref(self, parent_ref: V1ParentReference) -> None: ...



from kubernetes_asyncio.client.models import V1ObjectMeta
from kubernetes_asyncio.client.models import V1TopologySelectorTerm
from typing import Any

class V1StorageClass:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, allow_volume_expansion: bool | None = None, allowed_topologies: list[V1TopologySelectorTerm] | None = None, api_version: str | None = None, kind: str | None = None, metadata: V1ObjectMeta | None = None, mount_options: list[str] | None = None, parameters: dict[str, str] | None = None, provisioner: str | None = None, reclaim_policy: str | None = None, volume_binding_mode: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def allow_volume_expansion(self) -> bool: ...
    @allow_volume_expansion.setter
    def allow_volume_expansion(self, allow_volume_expansion: bool) -> None: ...
    @property
    def allowed_topologies(self) -> list[V1TopologySelectorTerm]: ...
    @allowed_topologies.setter
    def allowed_topologies(self, allowed_topologies: list[V1TopologySelectorTerm]) -> None: ...
    @property
    def api_version(self) -> str: ...
    @api_version.setter
    def api_version(self, api_version: str) -> None: ...
    @property
    def kind(self) -> str: ...
    @kind.setter
    def kind(self, kind: str) -> None: ...
    @property
    def metadata(self) -> V1ObjectMeta: ...
    @metadata.setter
    def metadata(self, metadata: V1ObjectMeta) -> None: ...
    @property
    def mount_options(self) -> list[str]: ...
    @mount_options.setter
    def mount_options(self, mount_options: list[str]) -> None: ...
    @property
    def parameters(self) -> dict[str, str]: ...
    @parameters.setter
    def parameters(self, parameters: dict[str, str]) -> None: ...
    @property
    def provisioner(self) -> str: ...
    @provisioner.setter
    def provisioner(self, provisioner: str) -> None: ...
    @property
    def reclaim_policy(self) -> str: ...
    @reclaim_policy.setter
    def reclaim_policy(self, reclaim_policy: str) -> None: ...
    @property
    def volume_binding_mode(self) -> str: ...
    @volume_binding_mode.setter
    def volume_binding_mode(self, volume_binding_mode: str) -> None: ...



from typing import Any

class V1TokenReviewSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, audiences: list[str] | None = None, token: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def audiences(self) -> list[str]: ...
    @audiences.setter
    def audiences(self, audiences: list[str]) -> None: ...
    @property
    def token(self) -> str: ...
    @token.setter
    def token(self, token: str) -> None: ...



from kubernetes_asyncio.client.models import V1ReplicaSetCondition
from typing import Any

class V1ReplicaSetStatus:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, available_replicas: int | None = None, conditions: list[V1ReplicaSetCondition] | None = None, fully_labeled_replicas: int | None = None, observed_generation: int | None = None, ready_replicas: int | None = None, replicas: int | None = None, terminating_replicas: int | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def available_replicas(self) -> int: ...
    @available_replicas.setter
    def available_replicas(self, available_replicas: int) -> None: ...
    @property
    def conditions(self) -> list[V1ReplicaSetCondition]: ...
    @conditions.setter
    def conditions(self, conditions: list[V1ReplicaSetCondition]) -> None: ...
    @property
    def fully_labeled_replicas(self) -> int: ...
    @fully_labeled_replicas.setter
    def fully_labeled_replicas(self, fully_labeled_replicas: int) -> None: ...
    @property
    def observed_generation(self) -> int: ...
    @observed_generation.setter
    def observed_generation(self, observed_generation: int) -> None: ...
    @property
    def ready_replicas(self) -> int: ...
    @ready_replicas.setter
    def ready_replicas(self, ready_replicas: int) -> None: ...
    @property
    def replicas(self) -> int: ...
    @replicas.setter
    def replicas(self, replicas: int) -> None: ...
    @property
    def terminating_replicas(self) -> int: ...
    @terminating_replicas.setter
    def terminating_replicas(self, terminating_replicas: int) -> None: ...



from kubernetes_asyncio.client.models import V1ContainerStateRunning
from kubernetes_asyncio.client.models import V1ContainerStateTerminated
from kubernetes_asyncio.client.models import V1ContainerStateWaiting
from typing import Any

class V1ContainerState:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, running: V1ContainerStateRunning | None = None, terminated: V1ContainerStateTerminated | None = None, waiting: V1ContainerStateWaiting | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def running(self) -> V1ContainerStateRunning: ...
    @running.setter
    def running(self, running: V1ContainerStateRunning) -> None: ...
    @property
    def terminated(self) -> V1ContainerStateTerminated: ...
    @terminated.setter
    def terminated(self, terminated: V1ContainerStateTerminated) -> None: ...
    @property
    def waiting(self) -> V1ContainerStateWaiting: ...
    @waiting.setter
    def waiting(self, waiting: V1ContainerStateWaiting) -> None: ...



from kubernetes_asyncio.client.models import V1beta1DeviceClassConfiguration
from kubernetes_asyncio.client.models import V1beta1DeviceSelector
from typing import Any

class V1beta1DeviceClassSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, config: list[V1beta1DeviceClassConfiguration] | None = None, extended_resource_name: str | None = None, selectors: list[V1beta1DeviceSelector] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def config(self) -> list[V1beta1DeviceClassConfiguration]: ...
    @config.setter
    def config(self, config: list[V1beta1DeviceClassConfiguration]) -> None: ...
    @property
    def extended_resource_name(self) -> str: ...
    @extended_resource_name.setter
    def extended_resource_name(self, extended_resource_name: str) -> None: ...
    @property
    def selectors(self) -> list[V1beta1DeviceSelector]: ...
    @selectors.setter
    def selectors(self, selectors: list[V1beta1DeviceSelector]) -> None: ...



from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.models import V1APIResourceList
from kubernetes_asyncio.client.models import V1CronJob
from kubernetes_asyncio.client.models import V1CronJobList
from kubernetes_asyncio.client.models import V1Job
from kubernetes_asyncio.client.models import V1JobList
from kubernetes_asyncio.client.models import V1Status
from multidict import CIMultiDictProxy
from typing import Any
from typing import Awaitable

class BatchV1Api:
    def __init__(self, api_client: ApiClient | None = None) -> None: ...
    def create_namespaced_cron_job(self, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def create_namespaced_cron_job_with_http_info(self, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def create_namespaced_job(self, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[V1Job]: ...
    def create_namespaced_job_with_http_info(self, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def delete_collection_namespaced_cron_job(self, namespace: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_collection_namespaced_cron_job_with_http_info(self, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def delete_collection_namespaced_job(self, namespace: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_collection_namespaced_job_with_http_info(self, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def delete_namespaced_cron_job(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_namespaced_cron_job_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def delete_namespaced_job(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_namespaced_job_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def get_api_resources(self, **kwargs: Any) -> Awaitable[V1APIResourceList]: ...
    def get_api_resources_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1APIResourceList, int, CIMultiDictProxy]]: ...
    def list_cron_job_for_all_namespaces(self, **kwargs: Any) -> Awaitable[V1CronJobList]: ...
    def list_cron_job_for_all_namespaces_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1CronJobList, int, CIMultiDictProxy]]: ...
    def list_job_for_all_namespaces(self, **kwargs: Any) -> Awaitable[V1JobList]: ...
    def list_job_for_all_namespaces_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1JobList, int, CIMultiDictProxy]]: ...
    def list_namespaced_cron_job(self, namespace: str, **kwargs: Any) -> Awaitable[V1CronJobList]: ...
    def list_namespaced_cron_job_with_http_info(self, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1CronJobList, int, CIMultiDictProxy]]: ...
    def list_namespaced_job(self, namespace: str, **kwargs: Any) -> Awaitable[V1JobList]: ...
    def list_namespaced_job_with_http_info(self, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1JobList, int, CIMultiDictProxy]]: ...
    def patch_namespaced_cron_job(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def patch_namespaced_cron_job_status(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def patch_namespaced_cron_job_status_with_http_info(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def patch_namespaced_cron_job_with_http_info(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def patch_namespaced_job(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[V1Job]: ...
    def patch_namespaced_job_status(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[V1Job]: ...
    def patch_namespaced_job_status_with_http_info(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def patch_namespaced_job_with_http_info(self, name: str, namespace: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def read_namespaced_cron_job(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def read_namespaced_cron_job_status(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def read_namespaced_cron_job_status_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def read_namespaced_cron_job_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def read_namespaced_job(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1Job]: ...
    def read_namespaced_job_status(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[V1Job]: ...
    def read_namespaced_job_status_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def read_namespaced_job_with_http_info(self, name: str, namespace: str, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def replace_namespaced_cron_job(self, name: str, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def replace_namespaced_cron_job_status(self, name: str, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[V1CronJob]: ...
    def replace_namespaced_cron_job_status_with_http_info(self, name: str, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def replace_namespaced_cron_job_with_http_info(self, name: str, namespace: str, body: V1CronJob, **kwargs: Any) -> Awaitable[tuple[V1CronJob, int, CIMultiDictProxy]]: ...
    def replace_namespaced_job(self, name: str, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[V1Job]: ...
    def replace_namespaced_job_status(self, name: str, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[V1Job]: ...
    def replace_namespaced_job_status_with_http_info(self, name: str, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...
    def replace_namespaced_job_with_http_info(self, name: str, namespace: str, body: V1Job, **kwargs: Any) -> Awaitable[tuple[V1Job, int, CIMultiDictProxy]]: ...



from kubernetes_asyncio.client.models import V1AppArmorProfile
from kubernetes_asyncio.client.models import V1Capabilities
from kubernetes_asyncio.client.models import V1SELinuxOptions
from kubernetes_asyncio.client.models import V1SeccompProfile
from kubernetes_asyncio.client.models import V1WindowsSecurityContextOptions
from typing import Any

class V1SecurityContext:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, allow_privilege_escalation: bool | None = None, app_armor_profile: V1AppArmorProfile | None = None, capabilities: V1Capabilities | None = None, privileged: bool | None = None, proc_mount: str | None = None, read_only_root_filesystem: bool | None = None, run_as_group: int | None = None, run_as_non_root: bool | None = None, run_as_user: int | None = None, se_linux_options: V1SELinuxOptions | None = None, seccomp_profile: V1SeccompProfile | None = None, windows_options: V1WindowsSecurityContextOptions | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def allow_privilege_escalation(self) -> bool: ...
    @allow_privilege_escalation.setter
    def allow_privilege_escalation(self, allow_privilege_escalation: bool) -> None: ...
    @property
    def app_armor_profile(self) -> V1AppArmorProfile: ...
    @app_armor_profile.setter
    def app_armor_profile(self, app_armor_profile: V1AppArmorProfile) -> None: ...
    @property
    def capabilities(self) -> V1Capabilities: ...
    @capabilities.setter
    def capabilities(self, capabilities: V1Capabilities) -> None: ...
    @property
    def privileged(self) -> bool: ...
    @privileged.setter
    def privileged(self, privileged: bool) -> None: ...
    @property
    def proc_mount(self) -> str: ...
    @proc_mount.setter
    def proc_mount(self, proc_mount: str) -> None: ...
    @property
    def read_only_root_filesystem(self) -> bool: ...
    @read_only_root_filesystem.setter
    def read_only_root_filesystem(self, read_only_root_filesystem: bool) -> None: ...
    @property
    def run_as_group(self) -> int: ...
    @run_as_group.setter
    def run_as_group(self, run_as_group: int) -> None: ...
    @property
    def run_as_non_root(self) -> bool: ...
    @run_as_non_root.setter
    def run_as_non_root(self, run_as_non_root: bool) -> None: ...
    @property
    def run_as_user(self) -> int: ...
    @run_as_user.setter
    def run_as_user(self, run_as_user: int) -> None: ...
    @property
    def se_linux_options(self) -> V1SELinuxOptions: ...
    @se_linux_options.setter
    def se_linux_options(self, se_linux_options: V1SELinuxOptions) -> None: ...
    @property
    def seccomp_profile(self) -> V1SeccompProfile: ...
    @seccomp_profile.setter
    def seccomp_profile(self, seccomp_profile: V1SeccompProfile) -> None: ...
    @property
    def windows_options(self) -> V1WindowsSecurityContextOptions: ...
    @windows_options.setter
    def windows_options(self, windows_options: V1WindowsSecurityContextOptions) -> None: ...



from datetime import datetime
from typing import Any

class V1ManagedFieldsEntry:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, api_version: str | None = None, fields_type: str | None = None, fields_v1: Any | None = None, manager: str | None = None, operation: str | None = None, subresource: str | None = None, time: datetime | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def api_version(self) -> str: ...
    @api_version.setter
    def api_version(self, api_version: str) -> None: ...
    @property
    def fields_type(self) -> str: ...
    @fields_type.setter
    def fields_type(self, fields_type: str) -> None: ...
    @property
    def fields_v1(self) -> Any: ...
    @fields_v1.setter
    def fields_v1(self, fields_v1: Any) -> None: ...
    @property
    def manager(self) -> str: ...
    @manager.setter
    def manager(self, manager: str) -> None: ...
    @property
    def operation(self) -> str: ...
    @operation.setter
    def operation(self, operation: str) -> None: ...
    @property
    def subresource(self) -> str: ...
    @subresource.setter
    def subresource(self, subresource: str) -> None: ...
    @property
    def time(self) -> datetime: ...
    @time.setter
    def time(self, time: datetime) -> None: ...



import asyncio
import json
from contextlib import contextmanager
from typing import Generator
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch

from aiohttp import web
from aiohttp.test_utils import (
    TestClient as _TestClient, TestServer as _TestServer,
)

from kubernetes_asyncio.config.config_exception import ConfigException
from kubernetes_asyncio.config.openid import OpenIDRequestor


def make_responder(response):
    async def responder(request):
        return response

    return responder


def respond_json(data):
    return make_responder(
        web.Response(
            text=json.dumps(data),
            content_type="application/json",
        )
    )


@contextmanager
def working_client() -> Generator[_TestClient, None, None]:
    loop = asyncio.get_event_loop()
    app = web.Application()
    app.router.add_get(
        "/.well-known/openid-configuration", respond_json({"token_endpoint": "/token"})
    )
    app.router.add_post(
        "/token",
        respond_json(
            {"id-token": "id-token-data", "refresh-token": "refresh-token-data"}
        ),
    )

    with patch(
        "kubernetes_asyncio.config.openid.aiohttp.ClientSession"
    ) as _client_session:
        client = _TestClient(_TestServer(app, loop=loop), loop=loop)
        _client_session.return_value = client

        yield client


@contextmanager
def fail_well_known_client() -> Generator[_TestClient, None, None]:
    loop = asyncio.get_event_loop()
    app = web.Application()
    app.router.add_get(
        "/.well-known/openid-configuration", make_responder(web.Response(status=500))
    )

    with patch(
        "kubernetes_asyncio.config.openid.aiohttp.ClientSession"
    ) as _client_session:
        client = _TestClient(_TestServer(app, loop=loop), loop=loop)
        _client_session.return_value = client
        yield client


@contextmanager
def fail_token_request_client() -> Generator[_TestClient, None, None]:
    loop = asyncio.get_event_loop()
    app = web.Application()
    app.router.add_get(
        "/.well-known/openid-configuration", respond_json({"token_endpoint": "/token"})
    )
    app.router.add_post("/token", make_responder(web.Response(status=500)))

    with patch(
        "kubernetes_asyncio.config.openid.aiohttp.ClientSession"
    ) as _client_session:
        client = _TestClient(_TestServer(app, loop=loop), loop=loop)
        _client_session.return_value = client

        yield client


class OpenIDRequestorTest(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.requestor = OpenIDRequestor(
            "client-id",
            "client-secret",
            "",
        )

    async def test_refresh_token_success(self) -> None:
        with working_client():
            resp = await self.requestor.refresh_token("my-refresh-token")

            assert resp["id-token"] == "id-token-data"
            assert resp["refresh-token"] == "refresh-token-data"

    async def test_failed_well_known(self) -> None:
        with fail_well_known_client():
            with self.assertRaises(ConfigException):
                await self.requestor.refresh_token("my-refresh-token")

    async def test_failed_refresh_token(self) -> None:
        with fail_token_request_client():
            with self.assertRaises(ConfigException):
                await self.requestor.refresh_token("my-refresh-token")



from datetime import datetime
from kubernetes_asyncio.client.models import V1ContainerStatus
from kubernetes_asyncio.client.models import V1HostIP
from kubernetes_asyncio.client.models import V1PodCondition
from kubernetes_asyncio.client.models import V1PodExtendedResourceClaimStatus
from kubernetes_asyncio.client.models import V1PodIP
from kubernetes_asyncio.client.models import V1PodResourceClaimStatus
from typing import Any

class V1PodStatus:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, conditions: list[V1PodCondition] | None = None, container_statuses: list[V1ContainerStatus] | None = None, ephemeral_container_statuses: list[V1ContainerStatus] | None = None, extended_resource_claim_status: V1PodExtendedResourceClaimStatus | None = None, host_ip: str | None = None, host_ips: list[V1HostIP] | None = None, init_container_statuses: list[V1ContainerStatus] | None = None, message: str | None = None, nominated_node_name: str | None = None, observed_generation: int | None = None, phase: str | None = None, pod_ip: str | None = None, pod_ips: list[V1PodIP] | None = None, qos_class: str | None = None, reason: str | None = None, resize: str | None = None, resource_claim_statuses: list[V1PodResourceClaimStatus] | None = None, start_time: datetime | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def conditions(self) -> list[V1PodCondition]: ...
    @conditions.setter
    def conditions(self, conditions: list[V1PodCondition]) -> None: ...
    @property
    def container_statuses(self) -> list[V1ContainerStatus]: ...
    @container_statuses.setter
    def container_statuses(self, container_statuses: list[V1ContainerStatus]) -> None: ...
    @property
    def ephemeral_container_statuses(self) -> list[V1ContainerStatus]: ...
    @ephemeral_container_statuses.setter
    def ephemeral_container_statuses(self, ephemeral_container_statuses: list[V1ContainerStatus]) -> None: ...
    @property
    def extended_resource_claim_status(self) -> V1PodExtendedResourceClaimStatus: ...
    @extended_resource_claim_status.setter
    def extended_resource_claim_status(self, extended_resource_claim_status: V1PodExtendedResourceClaimStatus) -> None: ...
    @property
    def host_ip(self) -> str: ...
    @host_ip.setter
    def host_ip(self, host_ip: str) -> None: ...
    @property
    def host_ips(self) -> list[V1HostIP]: ...
    @host_ips.setter
    def host_ips(self, host_ips: list[V1HostIP]) -> None: ...
    @property
    def init_container_statuses(self) -> list[V1ContainerStatus]: ...
    @init_container_statuses.setter
    def init_container_statuses(self, init_container_statuses: list[V1ContainerStatus]) -> None: ...
    @property
    def message(self) -> str: ...
    @message.setter
    def message(self, message: str) -> None: ...
    @property
    def nominated_node_name(self) -> str: ...
    @nominated_node_name.setter
    def nominated_node_name(self, nominated_node_name: str) -> None: ...
    @property
    def observed_generation(self) -> int: ...
    @observed_generation.setter
    def observed_generation(self, observed_generation: int) -> None: ...
    @property
    def phase(self) -> str: ...
    @phase.setter
    def phase(self, phase: str) -> None: ...
    @property
    def pod_ip(self) -> str: ...
    @pod_ip.setter
    def pod_ip(self, pod_ip: str) -> None: ...
    @property
    def pod_ips(self) -> list[V1PodIP]: ...
    @pod_ips.setter
    def pod_ips(self, pod_ips: list[V1PodIP]) -> None: ...
    @property
    def qos_class(self) -> str: ...
    @qos_class.setter
    def qos_class(self, qos_class: str) -> None: ...
    @property
    def reason(self) -> str: ...
    @reason.setter
    def reason(self, reason: str) -> None: ...
    @property
    def resize(self) -> str: ...
    @resize.setter
    def resize(self, resize: str) -> None: ...
    @property
    def resource_claim_statuses(self) -> list[V1PodResourceClaimStatus]: ...
    @resource_claim_statuses.setter
    def resource_claim_statuses(self, resource_claim_statuses: list[V1PodResourceClaimStatus]) -> None: ...
    @property
    def start_time(self) -> datetime: ...
    @start_time.setter
    def start_time(self, start_time: datetime) -> None: ...



from kubernetes_asyncio.client.models import V1NonResourceAttributes
from kubernetes_asyncio.client.models import V1ResourceAttributes
from typing import Any

class V1SubjectAccessReviewSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, extra: dict[str, list[str]] | None = None, groups: list[str] | None = None, non_resource_attributes: V1NonResourceAttributes | None = None, resource_attributes: V1ResourceAttributes | None = None, uid: str | None = None, user: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def extra(self) -> dict[str, list[str]]: ...
    @extra.setter
    def extra(self, extra: dict[str, list[str]]) -> None: ...
    @property
    def groups(self) -> list[str]: ...
    @groups.setter
    def groups(self, groups: list[str]) -> None: ...
    @property
    def non_resource_attributes(self) -> V1NonResourceAttributes: ...
    @non_resource_attributes.setter
    def non_resource_attributes(self, non_resource_attributes: V1NonResourceAttributes) -> None: ...
    @property
    def resource_attributes(self) -> V1ResourceAttributes: ...
    @resource_attributes.setter
    def resource_attributes(self, resource_attributes: V1ResourceAttributes) -> None: ...
    @property
    def uid(self) -> str: ...
    @uid.setter
    def uid(self, uid: str) -> None: ...
    @property
    def user(self) -> str: ...
    @user.setter
    def user(self, user: str) -> None: ...



from typing import Any

class V1beta2CapacityRequestPolicyRange:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, max: str | None = None, min: str | None = None, step: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def max(self) -> str: ...
    @max.setter
    def max(self, max: str) -> None: ...
    @property
    def min(self) -> str: ...
    @min.setter
    def min(self, min: str) -> None: ...
    @property
    def step(self) -> str: ...
    @step.setter
    def step(self, step: str) -> None: ...



from kubernetes_asyncio.client.models import V1LifecycleHandler
from typing import Any

class V1Lifecycle:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, post_start: V1LifecycleHandler | None = None, pre_stop: V1LifecycleHandler | None = None, stop_signal: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def post_start(self) -> V1LifecycleHandler: ...
    @post_start.setter
    def post_start(self, post_start: V1LifecycleHandler) -> None: ...
    @property
    def pre_stop(self) -> V1LifecycleHandler: ...
    @pre_stop.setter
    def pre_stop(self, pre_stop: V1LifecycleHandler) -> None: ...
    @property
    def stop_signal(self) -> str: ...
    @stop_signal.setter
    def stop_signal(self, stop_signal: str) -> None: ...



# Copyright 2019 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import logging
import os
import tempfile
from abc import abstractmethod
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING, Any, Generator

from aiohttp.client_exceptions import ContentTypeError
from urllib3.exceptions import MaxRetryError, ProtocolError

from kubernetes_asyncio import __version__

if TYPE_CHECKING:
    from kubernetes_asyncio.dynamic.client import DynamicClient

from kubernetes_asyncio.dynamic.exceptions import (
    NotFoundError, ResourceNotFoundError, ResourceNotUniqueError,
    ServiceUnavailableError,
)
from kubernetes_asyncio.dynamic.resource import Resource, ResourceList

DISCOVERY_PREFIX = "apis"
logger = logging.getLogger(__name__)


class ResourceGroup(object):
    """Helper class for Discoverer container"""

    def __init__(self, preferred, resources=None):
        self.preferred = preferred
        self.resources = resources or {}

    def to_dict(self):
        return {
            "_type": "ResourceGroup",
            "preferred": self.preferred,
            "resources": self.resources,
        }


class Discoverer(object):
    """
    A convenient container for storing discovered API resources. Allows
    easy searching and retrieval of specific resources.

    Subclasses implement the abstract methods with different loading strategies.
    """

    __version: dict[str, Any]

    def __init__(self, client: "DynamicClient", cache_file: str | None = None) -> None:
        self.client = client
        assert self.client.configuration.host
        default_cache_id = self.client.configuration.host.encode("utf-8")
        try:
            default_cachefile_name = "osrcp-{0}.json".format(
                hashlib.md5(default_cache_id, usedforsecurity=False).hexdigest()
            )
        except TypeError:
            # usedforsecurity is only supported in 3.9+
            default_cachefile_name = "osrcp-{0}.json".format(
                hashlib.md5(default_cache_id).hexdigest()
            )
        self.__cache_file = cache_file or os.path.join(
            tempfile.gettempdir(), default_cachefile_name
        )
        self._cache: dict[str, dict | str]

    def __await__(self) -> Generator[Any, None, "Discoverer"]:
        async def closure():
            await self.__init_cache()
            return self

        return closure().__await__()

    async def __aenter__(self) -> "Discoverer":
        await self.__init_cache()
        return self

    async def __init_cache(self, refresh: bool = False) -> None:
        if refresh or not os.path.exists(self.__cache_file):
            self._cache = {"library_version": __version__}
            refresh = True
        else:
            try:
                with open(self.__cache_file, "r") as f:
                    self._cache = json.load(f, cls=partial(CacheDecoder, self.client))  # type: ignore
                if self._cache.get("library_version") != __version__:
                    # Version mismatch, need to refresh cache
                    await self.invalidate_cache()
            except Exception as e:
                logger.error("load cache error: %s", e)
                await self.invalidate_cache()
        await self._load_server_info()
        await self.discover()
        if refresh:
            self._write_cache()

    def _write_cache(self) -> None:
        try:
            with open(self.__cache_file, "w") as f:
                json.dump(self._cache, f, cls=CacheEncoder)
        except Exception:
            # Failing to write the cache isn't a big enough error to crash on
            pass

    async def invalidate_cache(self) -> None:
        await self.__init_cache(refresh=True)

    @property
    @abstractmethod
    async def api_groups(self):
        pass

    @abstractmethod
    async def search(
        self,
        prefix: str | None = None,
        group: str | None = None,
        api_version: str | None = None,
        kind: str | None = None,
        **kwargs: Any,
    ) -> list[Resource]:
        pass

    @abstractmethod
    async def discover(self):
        pass

    @property
    def version(self) -> dict[str, Any]:
        return self.__version

    async def default_groups(
        self, request_resources: bool = False
    ) -> dict[str, dict[str, dict[str, ResourceGroup]]]:
        groups = {
            "api": {
                "": {
                    "v1": (
                        ResourceGroup(
                            True,
                            resources=await self.get_resources_for_api_version(
                                "api", "", "v1", True
                            ),
                        )
                        if request_resources
                        else ResourceGroup(True)
                    )
                }
            },
            DISCOVERY_PREFIX: {
                "": {
                    "v1": ResourceGroup(
                        True, resources={"List": [ResourceList(self.client)]}
                    )
                }
            },
        }
        return groups

    async def parse_api_groups(
        self, request_resources: bool = False, update: bool = False
    ) -> dict:
        """Discovers all API groups present in the cluster"""
        if not self._cache.get("resources") or update:
            self._cache["resources"] = self._cache.get("resources", {})
            response = await self.client.request("GET", "/{}".format(DISCOVERY_PREFIX))
            groups_response = response.groups

            groups = await self.default_groups(request_resources=request_resources)

            for group in groups_response:
                new_group = {}
                for version_raw in group["versions"]:
                    version = version_raw["version"]
                    resource_group = (
                        self._cache.get("resources", {})  # type: ignore
                        .get(DISCOVERY_PREFIX, {})  # type: ignore
                        .get(group["name"], {})
                        .get(version)
                    )
                    preferred = version_raw == group["preferredVersion"]
                    resources = resource_group.resources if resource_group else {}
                    if request_resources:
                        resources = await self.get_resources_for_api_version(
                            DISCOVERY_PREFIX, group["name"], version, preferred
                        )
                    new_group[version] = ResourceGroup(preferred, resources=resources)
                groups[DISCOVERY_PREFIX][group["name"]] = new_group
            self._cache["resources"].update(groups)  # type: ignore
            self._write_cache()

        return self._cache["resources"]  # type: ignore

    async def _load_server_info(self) -> None:
        def just_json(_, serialized):
            return serialized

        if not self._cache.get("version"):
            try:
                self._cache["version"] = {
                    "kubernetes": await self.client.request(
                        "get", "/version", serializer=just_json
                    )
                }
            except (ValueError, MaxRetryError) as e:
                if isinstance(e, MaxRetryError) and not isinstance(
                    e.reason, ProtocolError
                ):
                    raise
                if (
                    self.client.configuration.host is None
                    or not self.client.configuration.host.startswith("https://")
                ):
                    raise ValueError(
                        "Host value %s should start with https:// when talking to HTTPS endpoint"
                        % self.client.configuration.host
                    )
                else:
                    raise

        self.__version = self._cache["version"]  # type: ignore

    async def get_resources_for_api_version(
        self,
        prefix: str | None,
        group: str | None,
        version: str | None,
        preferred: bool,
    ) -> dict[str, list[Resource]]:
        """returns a dictionary of resources associated with provided (prefix, group, version)"""

        resources: dict[str, list[Resource]] = defaultdict(list)
        subresources: dict[str, dict] = {}

        path = "/".join(filter(None, [prefix, group, version]))
        try:
            response = await self.client.request("GET", path)
            resources_response = response.resources or []
        except (ServiceUnavailableError, ContentTypeError):
            # Handle both service unavailable errors and content type errors
            # (e.g., when server returns 503 with text/plain)
            resources_response = []

        resources_raw = list(filter(lambda r: "/" not in r["name"], resources_response))
        subresources_raw = list(filter(lambda r: "/" in r["name"], resources_response))
        for subresource in subresources_raw:
            # Handle resources with >2 parts in their name
            resource, name = subresource["name"].split("/", 1)
            if not subresources.get(resource):
                subresources[resource] = {}
            subresources[resource][name] = subresource

        for resource in resources_raw:
            # Prevent duplicate keys
            for key in ("prefix", "group", "api_version", "client", "preferred"):
                resource.pop(key, None)

            resourceobj = Resource(
                prefix=prefix,
                group=group,
                api_version=version,
                client=self.client,
                preferred=preferred,
                subresources=subresources.get(resource["name"]),
                **resource,
            )
            resources[resource["kind"]].append(resourceobj)

            resource_list = ResourceList(
                self.client,
                group=group,
                api_version=version,
                base_kind=resource["kind"],
            )
            resources[resource_list.kind].append(resource_list)  # type: ignore
        return resources

    async def get(self, **kwargs: Any) -> Resource:
        """Same as search, but will throw an error if there are multiple or no
        results. If there are multiple results and only one is an exact match
        on api_version, that resource will be returned.
        """
        results = await self.search(**kwargs)
        # If there are multiple matches, prefer exact matches on api_version
        if len(results) > 1 and kwargs.get("api_version"):
            results = [
                result
                for result in results
                if result.group_version == kwargs["api_version"]
            ]
        # If there are multiple matches, prefer non-List kinds
        if len(results) > 1 and not all([isinstance(x, ResourceList) for x in results]):
            results = [
                result for result in results if not isinstance(result, ResourceList)
            ]
        if len(results) == 1:
            return results[0]
        elif not results:
            raise ResourceNotFoundError("No matches found for {}".format(kwargs))
        else:
            raise ResourceNotUniqueError(
                "Multiple matches found for {}: {}".format(kwargs, results)
            )


class LazyDiscoverer(Discoverer):
    """A convenient container for storing discovered API resources. Allows
    easy searching and retrieval of specific resources.

    Resources for the cluster are loaded lazily.
    """

    def __init__(self, client, cache_file):
        self.__resources = None
        Discoverer.__init__(self, client, cache_file)
        self.__update_cache = False

    async def discover(self):
        self.__resources = await self.parse_api_groups(request_resources=False)

    def __maybe_write_cache(self):
        if self.__update_cache:
            self._write_cache()
            self.__update_cache = False

    @property
    async def api_groups(self):
        groups = await self.parse_api_groups(request_resources=False, update=True)
        return groups["apis"].keys()

    async def search(
        self,
        prefix: str | None = None,
        group: str | None = None,
        api_version: str | None = None,
        kind: str | None = None,
        **kwargs: Any,
    ) -> list[Resource]:
        # In first call, ignore ResourceNotFoundError and set default value for results
        try:
            results = await self.__search(
                self.__build_search(prefix, group, api_version, kind, **kwargs),
                self.__resources,
                [],
            )
        except ResourceNotFoundError:
            results = []
        if not results:
            await self.invalidate_cache()
            results = await self.__search(
                self.__build_search(prefix, group, api_version, kind, **kwargs),
                self.__resources,
                [],
            )
        self.__maybe_write_cache()
        return results

    async def __search(self, parts, resources, req_params):
        part = parts[0]
        if part != "*":
            resource_part = resources.get(part)
            if not resource_part:
                return []
            elif isinstance(resource_part, ResourceGroup):
                if len(req_params) != 2:
                    raise ValueError(
                        "prefix and group params should be present, have %s"
                        % req_params
                    )
                # Check if we've requested resources for this group
                if not resource_part.resources:
                    prefix, group, version = req_params[0], req_params[1], part
                    try:
                        resource_part.resources = (
                            await self.get_resources_for_api_version(
                                prefix, group, part, resource_part.preferred
                            )
                        )
                    except NotFoundError:
                        raise ResourceNotFoundError

                    self._cache["resources"][prefix][group][version] = resource_part  # type: ignore
                    self.__update_cache = True
                return await self.__search(
                    parts[1:], resource_part.resources, req_params
                )
            elif isinstance(resource_part, dict):
                # In this case parts [0] will be a specified prefix, group, version
                # as we recurse
                return await self.__search(
                    parts[1:], resource_part, req_params + [part]
                )
            else:
                if parts[1] != "*" and isinstance(parts[1], dict):
                    for _resource in resource_part:
                        for term, value in parts[1].items():
                            if getattr(_resource, term) == value:
                                return [_resource]
                    return []
                else:
                    return resource_part
        else:
            matches = []
            for key in resources.keys():
                matches.extend(
                    await self.__search([key] + parts[1:], resources, req_params)
                )
            return matches

    @staticmethod
    def __build_search(prefix=None, group=None, api_version=None, kind=None, **kwargs):
        if not group and api_version and "/" in api_version:
            group, api_version = api_version.split("/")

        items = [prefix, group, api_version, kind, kwargs]
        return list(map(lambda x: x or "*", items))

    async def __aiter__(self):
        assert self.__resources is not None
        for prefix, groups in self.__resources.items():
            for group, versions in groups.items():
                for version, rg in versions.items():
                    # Request resources for this groupVersion if we haven't yet
                    if not rg.resources:
                        rg.resources = await self.get_resources_for_api_version(
                            prefix, group, version, rg.preferred
                        )
                        self._cache["resources"][prefix][group][version] = rg  # type: ignore
                        self.__update_cache = True
                    for resource in rg.resources:
                        yield resource
        self.__maybe_write_cache()


class EagerDiscoverer(Discoverer):
    """A convenient container for storing discovered API resources. Allows
    easy searching and retrieval of specific resources.

    All resources are discovered for the cluster upon object instantiation.
    """

    def update(self, resources):
        self.__resources = resources

    def __init__(self, client, cache_file):
        self.__resources = None
        Discoverer.__init__(self, client, cache_file)

    async def discover(self):
        self.__resources = await self.parse_api_groups(request_resources=True)

    @property
    async def api_groups(self):
        """list available api groups"""
        groups = await self.parse_api_groups(request_resources=True, update=True)
        return groups["apis"].keys()

    async def search(
        self,
        prefix: str | None = None,
        group: str | None = None,
        api_version: str | None = None,
        kind: str | None = None,
        **kwargs: Any,
    ) -> list[Resource]:
        """Takes keyword arguments and returns matching resources. The search
        will happen in the following order:
            prefix: The api prefix for a resource, ie, /api, /oapi, /apis. Can usually be ignored
            group: The api group of a resource. Will also be extracted from api_version if it is present there
            api_version: The api version of a resource
            kind: The kind of the resource
            arbitrary arguments (see below), in random order

        The arbitrary arguments can be any valid attribute for an Resource object
        """
        results = self.__search(
            self.__build_search(prefix, group, api_version, kind, **kwargs),
            self.__resources,
        )
        if not results:
            await self.invalidate_cache()
            results = self.__search(
                self.__build_search(prefix, group, api_version, kind, **kwargs),
                self.__resources,
            )
        return results

    @staticmethod
    def __build_search(prefix=None, group=None, api_version=None, kind=None, **kwargs):
        if not group and api_version and "/" in api_version:
            group, api_version = api_version.split("/")

        items = [prefix, group, api_version, kind, kwargs]
        return list(map(lambda x: x or "*", items))

    def __search(self, parts, resources):
        part = parts[0]
        resource_part = resources.get(part)

        if part != "*" and resource_part:
            if isinstance(resource_part, ResourceGroup):
                return self.__search(parts[1:], resource_part.resources)
            elif isinstance(resource_part, dict):
                return self.__search(parts[1:], resource_part)
            else:
                if parts[1] != "*" and isinstance(parts[1], dict):
                    for _resource in resource_part:
                        for term, value in parts[1].items():
                            if getattr(_resource, term) == value:
                                return [_resource]
                    return []
                else:
                    return resource_part
        elif part == "*":
            matches = []
            for key in resources.keys():
                matches.extend(self.__search([key] + parts[1:], resources))
            return matches
        return []

    def __iter__(self):
        assert self.__resources is not None
        for _, groups in self.__resources.items():
            for _, versions in groups.items():
                for _, resources in versions.items():
                    for _, resource in resources.items():
                        yield resource


class CacheEncoder(json.JSONEncoder):
    def default(self, o):
        return o.to_dict()


class CacheDecoder(json.JSONDecoder):
    def __init__(self, client, *args, **kwargs):
        self.client = client
        json.JSONDecoder.__init__(self, object_hook=self._object_hook, *args, **kwargs)

    def _object_hook(self, obj):
        if "_type" not in obj:
            return obj
        _type = obj.pop("_type")
        if _type == "Resource":
            return Resource(client=self.client, **obj)
        elif _type == "ResourceList":
            return ResourceList(self.client, **obj)
        elif _type == "ResourceGroup":
            return ResourceGroup(
                obj["preferred"], resources=self._object_hook(obj["resources"])
            )
        return obj



from typing import Any

class V1FieldSelectorRequirement:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, key: str | None = None, operator: str | None = None, values: list[str] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def key(self) -> str: ...
    @key.setter
    def key(self, key: str) -> None: ...
    @property
    def operator(self) -> str: ...
    @operator.setter
    def operator(self, operator: str) -> None: ...
    @property
    def values(self) -> list[str]: ...
    @values.setter
    def values(self, values: list[str]) -> None: ...



# Copyright 2021 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
import json
import logging
import sys
import time
from http import HTTPStatus

from kubernetes_asyncio.client.exceptions import ApiException
from kubernetes_asyncio.leaderelection.electionconfig import Config
from kubernetes_asyncio.leaderelection.leaderelectionrecord import (
    LeaderElectionRecord,
)

"""
This package implements leader election using an annotation in a Kubernetes
object. The onstarted_leading coroutine is run as a task, which is cancelled if
the leader lock is obtained and then lost.

At first all candidates are considered followers. The one to create a lock or
update an existing lock first becomes the leader and remains so until it fails
to renew its lease.
"""


class LeaderElection:
    def __init__(self, election_config: Config) -> None:
        if election_config is None:
            sys.exit("argument config not passed")

        # Latest record observed in the created lock object
        self.observed_record: LeaderElectionRecord | None = None

        # The configuration set for this candidate
        self.election_config: Config = election_config

        # Latest update time of the lock
        self.observed_time_milliseconds = 0

    # Point of entry to Leader election
    async def run(self) -> None:
        # Try to create/ acquire a lock
        if await self.acquire():
            logging.info(
                "%s successfully acquired lease", self.election_config.lock.identity
            )

            onstarted_leading_coroutine = (
                self.election_config.onstarted_leading()
                if callable(self.election_config.onstarted_leading)
                else self.election_config.onstarted_leading
            )

            task = asyncio.create_task(onstarted_leading_coroutine)

            await self.renew_loop()

            # Leader lock lost - cancel the onstarted_leading coroutine if it's
            # still running. This permits onstarted_leading to clean up state
            # that might not be accessible to onstopped_leading.
            task.cancel()

            # Failed to update lease, run onstopped_leading callback. This is
            # preserved in order to continue to provide an interface similar to
            # the one provided by `kubernetes-client/python`.
            if self.election_config.onstopped_leading is not None:
                await (
                    self.election_config.onstopped_leading()
                    if callable(self.election_config.onstopped_leading)
                    else self.election_config.onstopped_leading
                )

    async def acquire(self) -> bool:
        # Follower
        logging.debug("%s is a follower", self.election_config.lock.identity)
        retry_period = self.election_config.retry_period

        while True:
            succeeded = await self.try_acquire_or_renew()

            if succeeded:
                return True

            await asyncio.sleep(retry_period)

    async def renew_loop(self) -> None:
        # Leader
        logging.debug(
            "Leader has entered renew loop and will try to update lease continuously"
        )

        retry_period = self.election_config.retry_period
        renew_deadline = self.election_config.renew_deadline * 1000

        while True:
            timeout = int(time.time() * 1000) + renew_deadline
            succeeded = False

            while int(time.time() * 1000) < timeout:
                succeeded = await self.try_acquire_or_renew()

                if succeeded:
                    break
                await asyncio.sleep(retry_period)

            if succeeded:
                await asyncio.sleep(retry_period)
                continue

            # failed to renew, return
            return

    async def try_acquire_or_renew(self) -> bool:
        now_timestamp = time.time()
        now = datetime.datetime.fromtimestamp(now_timestamp)

        # Check if lock is created
        lock_status, old_election_record = await self.election_config.lock.get(
            self.election_config.lock.name, self.election_config.lock.namespace
        )

        # create a default Election record for this candidate
        leader_election_record = LeaderElectionRecord(
            self.election_config.lock.identity,
            str(self.election_config.lease_duration),
            str(now),
            str(now),
        )

        # A lock is not created with that name, try to create one
        if not lock_status:
            assert (
                isinstance(old_election_record, ApiException)
                and old_election_record.body is not None
            )
            if json.loads(old_election_record.body)["code"] != HTTPStatus.NOT_FOUND:
                logging.error(
                    "Error retrieving resource lock %s as %s",
                    self.election_config.lock.name,
                    old_election_record.reason,
                )
                return False

            logging.debug(
                "%s is trying to create a lock",
                leader_election_record.holder_identity,
            )
            create_status = await self.election_config.lock.create(
                name=self.election_config.lock.name,
                namespace=self.election_config.lock.namespace,
                election_record=leader_election_record,
            )

            if not create_status:
                logging.error(
                    "%s failed to create lock", leader_election_record.holder_identity
                )
                return False

            self.observed_record = leader_election_record
            self.observed_time_milliseconds = int(time.time() * 1000)
            return True

        # A lock exists with that name
        # Validate old_election_record
        if old_election_record is None:
            # try to update lock with proper election record
            return await self.update_lock(leader_election_record)

        assert isinstance(old_election_record, LeaderElectionRecord)
        if (
            old_election_record.holder_identity is None
            or old_election_record.lease_duration is None
            or old_election_record.acquire_time is None
            or old_election_record.renew_time is None
        ):
            # try to update lock with proper election record
            return await self.update_lock(leader_election_record)

        # Report transitions
        if (
            self.observed_record
            and self.observed_record.holder_identity
            != old_election_record.holder_identity
        ):
            logging.debug(
                "Leader has switched to %s", old_election_record.holder_identity
            )

        if (
            self.observed_record is None
            or old_election_record.__dict__ != self.observed_record.__dict__
        ):
            self.observed_record = old_election_record
            self.observed_time_milliseconds = int(time.time() * 1000)

        # If This candidate is not the leader and lease duration is yet to finish
        if (
            self.election_config.lock.identity != self.observed_record.holder_identity
            and self.observed_time_milliseconds
            + self.election_config.lease_duration * 1000
            > int(now_timestamp * 1000)
        ):
            logging.debug(
                "Yet to finish lease_duration, lease held by %s and has not expired",
                old_election_record.holder_identity,
            )
            return False

        # If this candidate is the Leader
        if self.election_config.lock.identity == self.observed_record.holder_identity:
            # Leader updates renewTime, but keeps acquire_time unchanged
            leader_election_record.acquire_time = self.observed_record.acquire_time

        return await self.update_lock(leader_election_record)

    async def update_lock(self, leader_election_record: LeaderElectionRecord) -> bool:
        # Update object with latest election record
        update_status = await self.election_config.lock.update(
            self.election_config.lock.name,
            self.election_config.lock.namespace,
            leader_election_record,
        )

        if not update_status:
            logging.warning(
                "%s failed to acquire lease", leader_election_record.holder_identity
            )
            return False

        self.observed_record = leader_election_record
        self.observed_time_milliseconds = int(time.time() * 1000)
        logging.debug(
            "Leader %s has successfully updated lease",
            leader_election_record.holder_identity,
        )
        return True



import io
from logging import Logger
from typing import Any
from aiohttp import ClientResponse, ClientSession, ClientTimeout
from multidict import CIMultiDictProxy

from kubernetes_asyncio.client.exceptions import (
    ApiException as ApiException,
    ApiValueError as ApiValueError,
)

logger: Logger

class RESTResponse(io.IOBase):
    aiohttp_response: ClientResponse
    status: int
    reason: str
    data: bytes
    def __init__(self, resp: ClientResponse, data: bytes) -> None: ...
    def getheaders(self) -> CIMultiDictProxy: ...
    def getheader(self, name: str, default: str | None = None) -> str | None: ...

class RESTClientObject:
    server_hostname: str
    proxy: str
    proxy_headers: dict[str, str]
    pool_manager: ClientSession
    def __init__(
        self, configuration, pools_size: int = 4, maxsize: int | None = None
    ) -> None: ...
    async def close(self) -> None: ...
    async def request(
        self,
        method,
        url,
        query_params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        body: Any | None = None,
        post_params: dict[str, Any] | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def GET(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def HEAD(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def OPTIONS(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        post_params: dict[str, Any] | None = None,
        body: Any | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def DELETE(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        body: Any | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def POST(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        post_params: dict[str, Any] | None = None,
        body: Any | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def PUT(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        post_params: dict[str, Any] | None = None,
        body: Any | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...
    async def PATCH(
        self,
        url,
        headers: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        post_params: dict[str, Any] | None = None,
        body: Any | None = None,
        _preload_content: bool = True,
        _request_timeout: float | tuple[float] | ClientTimeout | None = None,
    ): ...



from kubernetes_asyncio.client.models import V1ObjectReference
from typing import Any

class V1StorageOSPersistentVolumeSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, fs_type: str | None = None, read_only: bool | None = None, secret_ref: V1ObjectReference | None = None, volume_name: str | None = None, volume_namespace: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def fs_type(self) -> str: ...
    @fs_type.setter
    def fs_type(self, fs_type: str) -> None: ...
    @property
    def read_only(self) -> bool: ...
    @read_only.setter
    def read_only(self, read_only: bool) -> None: ...
    @property
    def secret_ref(self) -> V1ObjectReference: ...
    @secret_ref.setter
    def secret_ref(self, secret_ref: V1ObjectReference) -> None: ...
    @property
    def volume_name(self) -> str: ...
    @volume_name.setter
    def volume_name(self, volume_name: str) -> None: ...
    @property
    def volume_namespace(self) -> str: ...
    @volume_namespace.setter
    def volume_namespace(self, volume_namespace: str) -> None: ...



from kubernetes_asyncio.client.models import ApiextensionsV1ServiceReference
from typing import Any

class ApiextensionsV1WebhookClientConfig:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, ca_bundle: str | None = None, service: ApiextensionsV1ServiceReference | None = None, url: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def ca_bundle(self) -> str: ...
    @ca_bundle.setter
    def ca_bundle(self, ca_bundle: str) -> None: ...
    @property
    def service(self) -> ApiextensionsV1ServiceReference: ...
    @service.setter
    def service(self, service: ApiextensionsV1ServiceReference) -> None: ...
    @property
    def url(self) -> str: ...
    @url.setter
    def url(self, url: str) -> None: ...



# Copyright 2017 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock, patch

from kubernetes_asyncio import client
from kubernetes_asyncio.stream import WsApiClient
from kubernetes_asyncio.stream.ws_client import WsResponse, get_websocket_url


class WsMock:
    def __init__(self) -> None:
        self.iter = 0

    def __aiter__(self) -> "WsMock":
        return self

    async def __aexit__(self, exc_type, exc, tb) -> "WsMock":
        return self

    async def __aenter__(self) -> "WsMock":
        return self

    async def __anext__(self) -> WsResponse:
        self.iter += 1
        if self.iter > 5:
            raise StopAsyncIteration
        return WsResponse(200, (chr(1) + "mock").encode("utf-8"))


class WSClientTest(IsolatedAsyncioTestCase):
    def test_websocket_client(self) -> None:
        for url, ws_url in [
            ("http://localhost/api", "ws://localhost/api"),
            ("https://localhost/api", "wss://localhost/api"),
            ("https://domain.com/api", "wss://domain.com/api"),
            ("https://api.domain.com/api", "wss://api.domain.com/api"),
            ("http://api.domain.com", "ws://api.domain.com"),
            ("https://api.domain.com", "wss://api.domain.com"),
            ("http://api.domain.com/", "ws://api.domain.com/"),
            ("https://api.domain.com/", "wss://api.domain.com/"),
        ]:
            self.assertEqual(get_websocket_url(url), ws_url)

    async def test_exec_ws(self) -> None:
        mock = Mock()
        mock.RESTClientObject.return_value.pool_manager = mock
        mock.ws_connect.return_value = WsMock()
        with patch("kubernetes_asyncio.client.api_client.rest", mock):
            api_client = WsApiClient()
            api_client.configuration.host = "https://localhost"
            ws = client.CoreV1Api(api_client=api_client)
            resp = ws.connect_get_namespaced_pod_exec(
                "pod",
                "namespace",
                command="mock-command",
                stderr=True,
                stdin=False,
                stdout=True,
                tty=False,
            )

            ret = await resp
            self.assertEqual(ret, "mock" * 5)
            mock.ws_connect.assert_called_once_with(
                "wss://localhost/api/v1/namespaces/namespace/pods/pod/exec?"
                "command=mock-command&stderr=True&stdin=False&stdout=True&tty=False",
                headers={
                    "sec-websocket-protocol": "v4.channel.k8s.io",
                    "Accept": "*/*",
                    "User-Agent": api_client.user_agent,
                },
                heartbeat=None,
            )

    async def test_exec_ws_with_heartbeat(self) -> None:
        mock = Mock()
        mock.RESTClientObject.return_value.pool_manager = mock
        mock.ws_connect.return_value = WsMock()
        with patch("kubernetes_asyncio.client.api_client.rest", mock):
            api_client = WsApiClient(heartbeat=30)
            api_client.configuration.host = "https://localhost"
            ws = client.CoreV1Api(api_client=api_client)
            resp = ws.connect_get_namespaced_pod_exec(
                "pod",
                "namespace",
                command="mock-command",
                stderr=True,
                stdin=False,
                stdout=True,
                tty=False,
            )

            ret = await resp
            self.assertEqual(ret, "mock" * 5)
            mock.ws_connect.assert_called_once_with(
                "wss://localhost/api/v1/namespaces/namespace/pods/pod/exec?"
                "command=mock-command&stderr=True&stdin=False&stdout=True&tty=False",
                headers={
                    "sec-websocket-protocol": "v4.channel.k8s.io",
                    "Accept": "*/*",
                    "User-Agent": api_client.user_agent,
                },
                heartbeat=30,
            )

    def test_parse_error_data_success(self) -> None:
        error_data = '{"metadata":{},"status":"Success"}'
        return_code = WsApiClient.parse_error_data(error_data)
        self.assertEqual(return_code, 0)

    def test_parse_error_data_failure(self) -> None:
        error_data = (
            '{"metadata":{},"status":"Failure",'
            '"message":"command terminated with non-zero exit code",'
            '"reason":"NonZeroExitCode",'
            '"details":{"causes":[{"reason":"ExitCode","message":"1"}]}}'
        )
        return_code = WsApiClient.parse_error_data(error_data)
        self.assertEqual(return_code, 1)



# Copyright 2021 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Coroutine
from typing import Any

from kubernetes_asyncio.leaderelection.resourcelock.baselock import BaseLock


class Config:
    # Validate config, exit if an error is detected

    # onstarted_leading and onstopped_leading accept either coroutines or
    # coroutine functions. Coroutines faciliate passing context, but coroutine
    # functions can be simpler when passing context is not required.
    #
    # One example of when passing context is helpful is sharing the ApiClient
    # used by the leader election, which can then be used for subsequent
    # Kubernetes API operations upon onstopped_leading or onstopped_leading.
    def __init__(
        self,
        lock: BaseLock,
        lease_duration: float,
        renew_deadline: float,
        retry_period: float,
        onstarted_leading: Callable[[], Coroutine[Any, Any, None]]
        | Coroutine[Any, Any, None],
        onstopped_leading: Callable[[], Coroutine[Any, Any, None]]
        | Coroutine[Any, Any, None],
    ) -> None:
        self.jitter_factor = 1.2

        if lock is None:
            raise ValueError("lock cannot be None")
        self.lock = lock

        if lease_duration <= renew_deadline:
            raise ValueError("lease_duration must be greater than renew_deadline")

        if renew_deadline <= self.jitter_factor * retry_period:
            raise ValueError(
                "renewDeadline must be greater than retry_period*jitter_factor"
            )

        if lease_duration < 1:
            raise ValueError("lease_duration must be greater than one")

        if renew_deadline < 1:
            raise ValueError("renew_deadline must be greater than one")

        if retry_period < 1:
            raise ValueError("retry_period must be greater than one")

        self.lease_duration = lease_duration
        self.renew_deadline = renew_deadline
        self.retry_period = retry_period

        if onstarted_leading is None:
            raise ValueError("callback onstarted_leading cannot be None")
        self.onstarted_leading = onstarted_leading

        self.onstopped_leading = onstopped_leading



from kubernetes_asyncio.client.models import V1Preconditions
from typing import Any

class V1DeleteOptions:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, api_version: str | None = None, dry_run: list[str] | None = None, grace_period_seconds: int | None = None, ignore_store_read_error_with_cluster_breaking_potential: bool | None = None, kind: str | None = None, orphan_dependents: bool | None = None, preconditions: V1Preconditions | None = None, propagation_policy: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def api_version(self) -> str: ...
    @api_version.setter
    def api_version(self, api_version: str) -> None: ...
    @property
    def dry_run(self) -> list[str]: ...
    @dry_run.setter
    def dry_run(self, dry_run: list[str]) -> None: ...
    @property
    def grace_period_seconds(self) -> int: ...
    @grace_period_seconds.setter
    def grace_period_seconds(self, grace_period_seconds: int) -> None: ...
    @property
    def ignore_store_read_error_with_cluster_breaking_potential(self) -> bool: ...
    @ignore_store_read_error_with_cluster_breaking_potential.setter
    def ignore_store_read_error_with_cluster_breaking_potential(self, ignore_store_read_error_with_cluster_breaking_potential: bool) -> None: ...
    @property
    def kind(self) -> str: ...
    @kind.setter
    def kind(self, kind: str) -> None: ...
    @property
    def orphan_dependents(self) -> bool: ...
    @orphan_dependents.setter
    def orphan_dependents(self, orphan_dependents: bool) -> None: ...
    @property
    def preconditions(self) -> V1Preconditions: ...
    @preconditions.setter
    def preconditions(self, preconditions: V1Preconditions) -> None: ...
    @property
    def propagation_policy(self) -> str: ...
    @propagation_policy.setter
    def propagation_policy(self, propagation_policy: str) -> None: ...



from kubernetes_asyncio.client.models import V1FieldSelectorAttributes
from kubernetes_asyncio.client.models import V1LabelSelectorAttributes
from typing import Any

class V1ResourceAttributes:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, field_selector: V1FieldSelectorAttributes | None = None, group: str | None = None, label_selector: V1LabelSelectorAttributes | None = None, name: str | None = None, namespace: str | None = None, resource: str | None = None, subresource: str | None = None, verb: str | None = None, version: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def field_selector(self) -> V1FieldSelectorAttributes: ...
    @field_selector.setter
    def field_selector(self, field_selector: V1FieldSelectorAttributes) -> None: ...
    @property
    def group(self) -> str: ...
    @group.setter
    def group(self, group: str) -> None: ...
    @property
    def label_selector(self) -> V1LabelSelectorAttributes: ...
    @label_selector.setter
    def label_selector(self, label_selector: V1LabelSelectorAttributes) -> None: ...
    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, name: str) -> None: ...
    @property
    def namespace(self) -> str: ...
    @namespace.setter
    def namespace(self, namespace: str) -> None: ...
    @property
    def resource(self) -> str: ...
    @resource.setter
    def resource(self, resource: str) -> None: ...
    @property
    def subresource(self) -> str: ...
    @subresource.setter
    def subresource(self, subresource: str) -> None: ...
    @property
    def verb(self) -> str: ...
    @verb.setter
    def verb(self, verb: str) -> None: ...
    @property
    def version(self) -> str: ...
    @version.setter
    def version(self, version: str) -> None: ...



from kubernetes_asyncio.client.models import V1JobTemplateSpec
from typing import Any

class V1CronJobSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, concurrency_policy: str | None = None, failed_jobs_history_limit: int | None = None, job_template: V1JobTemplateSpec | None = None, schedule: str | None = None, starting_deadline_seconds: int | None = None, successful_jobs_history_limit: int | None = None, suspend: bool | None = None, time_zone: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def concurrency_policy(self) -> str: ...
    @concurrency_policy.setter
    def concurrency_policy(self, concurrency_policy: str) -> None: ...
    @property
    def failed_jobs_history_limit(self) -> int: ...
    @failed_jobs_history_limit.setter
    def failed_jobs_history_limit(self, failed_jobs_history_limit: int) -> None: ...
    @property
    def job_template(self) -> V1JobTemplateSpec: ...
    @job_template.setter
    def job_template(self, job_template: V1JobTemplateSpec) -> None: ...
    @property
    def schedule(self) -> str: ...
    @schedule.setter
    def schedule(self, schedule: str) -> None: ...
    @property
    def starting_deadline_seconds(self) -> int: ...
    @starting_deadline_seconds.setter
    def starting_deadline_seconds(self, starting_deadline_seconds: int) -> None: ...
    @property
    def successful_jobs_history_limit(self) -> int: ...
    @successful_jobs_history_limit.setter
    def successful_jobs_history_limit(self, successful_jobs_history_limit: int) -> None: ...
    @property
    def suspend(self) -> bool: ...
    @suspend.setter
    def suspend(self, suspend: bool) -> None: ...
    @property
    def time_zone(self) -> str: ...
    @time_zone.setter
    def time_zone(self, time_zone: str) -> None: ...



from kubernetes_asyncio.client.models import V1beta2DeviceToleration
from typing import Any

class V1beta2DeviceRequestAllocationResult:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, admin_access: bool | None = None, binding_conditions: list[str] | None = None, binding_failure_conditions: list[str] | None = None, consumed_capacity: dict[str, str] | None = None, device: str | None = None, driver: str | None = None, pool: str | None = None, request: str | None = None, share_id: str | None = None, tolerations: list[V1beta2DeviceToleration] | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def admin_access(self) -> bool: ...
    @admin_access.setter
    def admin_access(self, admin_access: bool) -> None: ...
    @property
    def binding_conditions(self) -> list[str]: ...
    @binding_conditions.setter
    def binding_conditions(self, binding_conditions: list[str]) -> None: ...
    @property
    def binding_failure_conditions(self) -> list[str]: ...
    @binding_failure_conditions.setter
    def binding_failure_conditions(self, binding_failure_conditions: list[str]) -> None: ...
    @property
    def consumed_capacity(self) -> dict[str, str]: ...
    @consumed_capacity.setter
    def consumed_capacity(self, consumed_capacity: dict[str, str]) -> None: ...
    @property
    def device(self) -> str: ...
    @device.setter
    def device(self, device: str) -> None: ...
    @property
    def driver(self) -> str: ...
    @driver.setter
    def driver(self, driver: str) -> None: ...
    @property
    def pool(self) -> str: ...
    @pool.setter
    def pool(self, pool: str) -> None: ...
    @property
    def request(self) -> str: ...
    @request.setter
    def request(self, request: str) -> None: ...
    @property
    def share_id(self) -> str: ...
    @share_id.setter
    def share_id(self, share_id: str) -> None: ...
    @property
    def tolerations(self) -> list[V1beta2DeviceToleration]: ...
    @tolerations.setter
    def tolerations(self, tolerations: list[V1beta2DeviceToleration]) -> None: ...



from datetime import datetime
from typing import Any

class V1alpha2LeaseCandidateSpec:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, binary_version: str | None = None, emulation_version: str | None = None, lease_name: str | None = None, ping_time: datetime | None = None, renew_time: datetime | None = None, strategy: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def binary_version(self) -> str: ...
    @binary_version.setter
    def binary_version(self, binary_version: str) -> None: ...
    @property
    def emulation_version(self) -> str: ...
    @emulation_version.setter
    def emulation_version(self, emulation_version: str) -> None: ...
    @property
    def lease_name(self) -> str: ...
    @lease_name.setter
    def lease_name(self, lease_name: str) -> None: ...
    @property
    def ping_time(self) -> datetime: ...
    @ping_time.setter
    def ping_time(self, ping_time: datetime) -> None: ...
    @property
    def renew_time(self) -> datetime: ...
    @renew_time.setter
    def renew_time(self, renew_time: datetime) -> None: ...
    @property
    def strategy(self) -> str: ...
    @strategy.setter
    def strategy(self, strategy: str) -> None: ...



from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.models import V1APIResourceList
from kubernetes_asyncio.client.models import V1APIService
from kubernetes_asyncio.client.models import V1APIServiceList
from kubernetes_asyncio.client.models import V1Status
from multidict import CIMultiDictProxy
from typing import Any
from typing import Awaitable

class ApiregistrationV1Api:
    def __init__(self, api_client: ApiClient | None = None) -> None: ...
    def create_api_service(self, body: V1APIService, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def create_api_service_with_http_info(self, body: V1APIService, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def delete_api_service(self, name: str, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_api_service_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def delete_collection_api_service(self, **kwargs: Any) -> Awaitable[V1Status]: ...
    def delete_collection_api_service_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1Status, int, CIMultiDictProxy]]: ...
    def get_api_resources(self, **kwargs: Any) -> Awaitable[V1APIResourceList]: ...
    def get_api_resources_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1APIResourceList, int, CIMultiDictProxy]]: ...
    def list_api_service(self, **kwargs: Any) -> Awaitable[V1APIServiceList]: ...
    def list_api_service_with_http_info(self, **kwargs: Any) -> Awaitable[tuple[V1APIServiceList, int, CIMultiDictProxy]]: ...
    def patch_api_service(self, name: str, body: Any, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def patch_api_service_status(self, name: str, body: Any, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def patch_api_service_status_with_http_info(self, name: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def patch_api_service_with_http_info(self, name: str, body: Any, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def read_api_service(self, name: str, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def read_api_service_status(self, name: str, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def read_api_service_status_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def read_api_service_with_http_info(self, name: str, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def replace_api_service(self, name: str, body: V1APIService, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def replace_api_service_status(self, name: str, body: V1APIService, **kwargs: Any) -> Awaitable[V1APIService]: ...
    def replace_api_service_status_with_http_info(self, name: str, body: V1APIService, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...
    def replace_api_service_with_http_info(self, name: str, body: V1APIService, **kwargs: Any) -> Awaitable[tuple[V1APIService, int, CIMultiDictProxy]]: ...



from typing import Any

class V1NonResourceAttributes:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, path: str | None = None, verb: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def path(self) -> str: ...
    @path.setter
    def path(self, path: str) -> None: ...
    @property
    def verb(self) -> str: ...
    @verb.setter
    def verb(self, verb: str) -> None: ...



from kubernetes_asyncio.client.models import V1LocalObjectReference
from typing import Any

class V1ISCSIVolumeSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, chap_auth_discovery: bool | None = None, chap_auth_session: bool | None = None, fs_type: str | None = None, initiator_name: str | None = None, iqn: str | None = None, iscsi_interface: str | None = None, lun: int | None = None, portals: list[str] | None = None, read_only: bool | None = None, secret_ref: V1LocalObjectReference | None = None, target_portal: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def chap_auth_discovery(self) -> bool: ...
    @chap_auth_discovery.setter
    def chap_auth_discovery(self, chap_auth_discovery: bool) -> None: ...
    @property
    def chap_auth_session(self) -> bool: ...
    @chap_auth_session.setter
    def chap_auth_session(self, chap_auth_session: bool) -> None: ...
    @property
    def fs_type(self) -> str: ...
    @fs_type.setter
    def fs_type(self, fs_type: str) -> None: ...
    @property
    def initiator_name(self) -> str: ...
    @initiator_name.setter
    def initiator_name(self, initiator_name: str) -> None: ...
    @property
    def iqn(self) -> str: ...
    @iqn.setter
    def iqn(self, iqn: str) -> None: ...
    @property
    def iscsi_interface(self) -> str: ...
    @iscsi_interface.setter
    def iscsi_interface(self, iscsi_interface: str) -> None: ...
    @property
    def lun(self) -> int: ...
    @lun.setter
    def lun(self, lun: int) -> None: ...
    @property
    def portals(self) -> list[str]: ...
    @portals.setter
    def portals(self, portals: list[str]) -> None: ...
    @property
    def read_only(self) -> bool: ...
    @read_only.setter
    def read_only(self, read_only: bool) -> None: ...
    @property
    def secret_ref(self) -> V1LocalObjectReference: ...
    @secret_ref.setter
    def secret_ref(self, secret_ref: V1LocalObjectReference) -> None: ...
    @property
    def target_portal(self) -> str: ...
    @target_portal.setter
    def target_portal(self, target_portal: str) -> None: ...



from typing import Any

class V1APIResource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, categories: list[str] | None = None, group: str | None = None, kind: str | None = None, name: str | None = None, namespaced: bool | None = None, short_names: list[str] | None = None, singular_name: str | None = None, storage_version_hash: str | None = None, verbs: list[str] | None = None, version: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def categories(self) -> list[str]: ...
    @categories.setter
    def categories(self, categories: list[str]) -> None: ...
    @property
    def group(self) -> str: ...
    @group.setter
    def group(self, group: str) -> None: ...
    @property
    def kind(self) -> str: ...
    @kind.setter
    def kind(self, kind: str) -> None: ...
    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, name: str) -> None: ...
    @property
    def namespaced(self) -> bool: ...
    @namespaced.setter
    def namespaced(self, namespaced: bool) -> None: ...
    @property
    def short_names(self) -> list[str]: ...
    @short_names.setter
    def short_names(self, short_names: list[str]) -> None: ...
    @property
    def singular_name(self) -> str: ...
    @singular_name.setter
    def singular_name(self, singular_name: str) -> None: ...
    @property
    def storage_version_hash(self) -> str: ...
    @storage_version_hash.setter
    def storage_version_hash(self, storage_version_hash: str) -> None: ...
    @property
    def verbs(self) -> list[str]: ...
    @verbs.setter
    def verbs(self, verbs: list[str]) -> None: ...
    @property
    def version(self) -> str: ...
    @version.setter
    def version(self, version: str) -> None: ...



from kubernetes_asyncio.client.models import V1SecretReference
from typing import Any

class V1CephFSPersistentVolumeSource:
    def __eq__(self, other: Any) -> bool: ...
    def __init__(self, monitors: list[str] | None = None, path: str | None = None, read_only: bool | None = None, secret_file: str | None = None, secret_ref: V1SecretReference | None = None, user: str | None = None, local_vars_configuration: Any | None = None) -> None: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def to_dict(self, serialize: bool = False) -> dict[str, Any]: ...
    def to_str(self) -> str: ...
    @property
    def monitors(self) -> list[str]: ...
    @monitors.setter
    def monitors(self, monitors: list[str]) -> None: ...
    @property
    def path(self) -> str: ...
    @path.setter
    def path(self, path: str) -> None: ...
    @property
    def read_only(self) -> bool: ...
    @read_only.setter
    def read_only(self, read_only: bool) -> None: ...
    @property
    def secret_file(self) -> str: ...
    @secret_file.setter
    def secret_file(self, secret_file: str) -> None: ...
    @property
    def secret_ref(self) -> V1SecretReference: ...
    @secret_ref.setter
    def secret_ref(self, secret_ref: V1SecretReference) -> None: ...
    @property
    def user(self) -> str: ...
    @user.setter
    def user(self, user: str) -> None: ...
